import logging
from pathlib import Path
import pytest

logger = logging.getLogger(__name__)

from alexber.utils.mains import fixabscwd, logger as real_logger
import os
import sys


_real_info = real_logger.info

@pytest.fixture(scope='module')
def loggingCaptureWarnings():
    logging.captureWarnings(True)

@pytest.fixture
def mock_os(mocker):
    ret_mock = mocker.patch('.'.join(['alexber.utils.mains', '_os']), autospec=True, spec_set=True)

    cwd = str(Path(__file__).resolve().parent)
    ret_mock.getcwd.return_value = cwd
    ret_mock.path = os.path

    return ret_mock

@pytest.fixture
def mock_info(mocker):
    ret_mock = mocker.patch.object(real_logger, 'info', side_effect=_real_info, autospec=True, spec_set=True)
    return ret_mock

@pytest.fixture
def mock_main_module(mocker):
    ret_mock = sys.modules['__main__']
    mocker.patch.object(ret_mock, '__file__', new=__file__)
    return ret_mock


def test_fixabscwd(request, loggingCaptureWarnings, mock_os, mock_info, mock_main_module):
    logger.info(f'{request._pyfuncitem.name}()')

    fixabscwd()

    pytest.assume(mock_info.call_count > 0)

    pytest.assume(mock_os.chdir.call_count == 1)
    (newcwd,), _ = mock_os.chdir.call_args
    pytest.assume(str(Path(__file__).parent)==newcwd)




if __name__ == "__main__":
    pytest.main([__file__])



import inspect
#from inspect import signature as inspect_signature
import logging
logger = logging.getLogger(__name__)


def issetdescriptor(object):
    """Return true if the object is a method descriptor with setters.

    But not if ismethod() or isclass() or isfunction() are true.
    """
    if inspect.isclass(object) or inspect.ismethod(object) or inspect.isfunction(object):
        # mutual exclusion
        return False
    tp = type(object)

    return hasattr(tp, "__set__")


def ismethod(object):
    '''
    If object is class, return false.
    If object is not function, return false.

    :param object:
    :return: false if object is not a class and not a function. Otherwise, return true iff signature has 2 params.
    '''
    if inspect.isclass(object):
        return False

    if not inspect.isfunction(object):
        return False


    return True

def has_method(cls, methodName):
    '''
    Check if class cls has method with name methodName directly,
    or in one of it's super-classes.

    :param cls:
    :param methodName:
    :return:
    '''

    #see https://mail.python.org/pipermail/python-dev/2010-August/103200.html
    #see https://docs.python.org/3.7/library/abc.html

    #logger.info(list(methodName in B.__dict__ for B in cls.__mro__))

    if any(methodName in B.__dict__ for B in cls.__mro__):
        return True
    return False



import logging

import os as _os
import sys as _sys
from collections import deque
from .parsers import is_empty
from .importer import importer

logger = logging.getLogger(__name__)


def fixabscwd():
    """
    See here https://medium.com/@alex_ber/making-relative-path-to-file-to-work-d5d0f1da67bf for documentation.
    """
    logger.info(f"cwd is {_os.getcwd()}")

    main_module = _sys.modules['__main__']

    #inspider by https://github.com/theskumar/python-dotenv/blob/12439aac2d7c98b5acaf51ae3044b1659dc086ae/src/dotenv/main.py#L250
    def _is_interactive():
        """ Decide whether this is running in a REPL or IPython notebook """
        return not hasattr(main_module, '__file__')

    if _is_interactive() or getattr(_sys, 'frozen', False):
        # Should work without __file__, e.g. in REPL or IPython notebook.
        pass
    else:
        main_dir = _os.path.dirname(_os.path.realpath(main_module.__file__))

        logger.info(f"Going to change os.chdir('{main_dir}')")

        _os.chdir(main_dir)

def path():
    """
    For older Python version uses`importlib_resources` module.
    For newer version built in `importlib.resources`.
    :return:
    """
    if _sys.version_info >= (3, 7):
        from importlib.resources import path as _path
    else:
        try:
            from importlib_resources import path as _path
        except ImportError:
            import warnings

            warning = (
                "You appear to be missing some optional dependencies (importlib_resources);"
            )
            warnings.warn(warning, ImportWarning)
            raise
    return _path


def load_env(**kwargs):
    """
    Convenient method from loading environment variables
    inside packed format (eggs, etc).

    if dotenv_path or stream is present it will be used.
    if ENV_PCK is present, dotenv_path will be constructed from ENV_PCK and ENV_NAME.
    Otherwise, kwargs will be forwarded as is to load_dotenv.


    Implementaion note:
    If available it uses importlib.resources API,
    if not it assumes existence of backport of importlib_resources.

    :param ENV_PCK: package where to find .env file Optional.
    :param ENV_NAME: Name of .env file. Optional.
    :param dotenv_path:  absolute or relative path to .env file. Optional.
    :param stream: `StringIO` object with .env content. Optional.
    :return:
    """

    # os.environ['AWS_DEFAULT_REGION'] = 'eu-central-1'
    try:
        from dotenv import load_dotenv
    except ImportError:
        import warnings


        warning = (
            "You appear to be missing some optional dependencies (python-dotenv);"
            "please 'python3 -m pip install alex-ber-utils[python-dotenv]'."
        )

        warnings.warn(warning, ImportWarning)
        raise

    l_path = path()

    dotenv_path = kwargs.get('dotenv_path', None)
    stream = kwargs.get('stream', None)
    ENV_PCK = kwargs.pop('ENV_PCK', None)

    if not is_empty(dotenv_path) or not is_empty(stream) or \
            (is_empty(dotenv_path) and is_empty(stream) and is_empty(ENV_PCK)):
        load_dotenv(**kwargs)

    else:
        if is_empty(ENV_PCK):
            raise ValueError("ENV_PCK can't be empty")

        ENV_NAME = kwargs.pop('ENV_NAME', '.env')
        with l_path(ENV_PCK, ENV_NAME) as full_path:
            d = {**kwargs, 'dotenv_path': full_path}

            load_dotenv(**d)



class BaseOsEnvrion:
    DEFAULT_DELIMSEP = _os.pathsep  # ';'
    DEFAULT_ENVSEP = _os.path.sep  # /
    DEFAULT_KEYSEP = ','

    def __init__(self, **kwargs):
        delimsep = kwargs.get('ENV_DELIM_SEP', None)
        if is_empty(delimsep):
            delimsep = self.DEFAULT_DELIMSEP
        self.delimsep = delimsep

        envsep = kwargs.get('ENV_SEP', None)
        if is_empty(envsep):
            envsep = self.DEFAULT_ENVSEP
        self.envsep = envsep


        keysep = kwargs.get('ENV_KEY_SEP', None)
        if is_empty(keysep):
            keysep = self.DEFAULT_KEYSEP
        self.keysep = keysep

        env_keys = kwargs.get('ENV_KEYS', None) # 'KEY'
        if is_empty(env_keys):
            raise ValueError('ENV_KEYS is not found in kwargs')
        self.env_keys = self._str_to_list(self.keysep, env_keys)

    def _str_to_list(self, sep, value):
        ret = [w.strip() for w in value.split(sep)]
        return ret

    def _list_to_str(self, sep, *args):
        ret = sep.join([*args])
        return ret


class OsEnvrionPathExpender(BaseOsEnvrion):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        pck = kwargs.get('ENV_MAIN_PCK', None)
        if is_empty(pck):
            raise ValueError('ENV_MAIN_PCK is not found in kwargs')

        self.pck = pck

    def fix_abs_path(self):
        l_path = path()

        with l_path(self.pck, '__init__.py') as init_file:
            full_prefix = str(init_file.parent.parent)

            for env_key in self.env_keys:
                rel_paths = self._str_to_list(self.delimsep, _os.environ[env_key])

                buffersize = 0 if rel_paths is None else len(rel_paths)
                env_values = deque(maxlen=buffersize)

                for rel_path in rel_paths:
                    abs_path = self._list_to_str(self.envsep, *[full_prefix, rel_path])
                    env_values.append(abs_path)

                values = self._list_to_str(self.delimsep, *env_values)
                _os.environ[env_key] = values


class OsEnvrionPathRetry(BaseOsEnvrion):

    def fix_retry_path(self):
        from pathlib import Path

        for env_key in self.env_keys:
            env_paths = self._str_to_list(self.delimsep, _os.environ[env_key])

            buffersize = 0 if env_paths is None else len(env_paths)
            env_values = deque(maxlen=buffersize)

            for env_path in env_paths:
                env_path_p = Path(env_path)
                is_exists = env_path_p.exists()

                if not is_exists:
                    drive = env_path_p.drive
                    if len(drive) == 2 and drive[1] == ':':
                        env_path_p = env_path_p.as_posix()[2:]

                env_values.append(env_path_p)

            values = self._list_to_str(self.delimsep, *env_values)
            _os.environ[env_key] = values



def fix_env(**kwargs):
    """
    This method "fixes" os.environ relatively to the given ENV_MAIN_PCK package.
    This method ignores current working directory or __main__ module's __file__ atrribute.
    For each key in ENV_KEYS, this method prepends full_prefix to os.environ[key].
    full_prefix is calculated as absolute path of __init__.py of ENV_MAIN_PCK.



    :param ENV_KEYS keys of os.environ which will be fixed.
    :param ENV_MAIN_PCK: package to calcualte full_prefix.
    :param ENV_KEY_SEP: seperator used in ENV_PCK. Optional. Default is ','
    :param ENV_SEP: seperator that is used inside path. Optional. Default is os.path.sep.
    :param ENV_DELIM_SEP: seperator that is used inside os.environ. Optional. Default is os.pathsep.
    :param cls: class or str with implementation logic. Optional. Default is OsEnvrionPathExpender.
    """

    cls = kwargs.pop('cls', OsEnvrionPathExpender)
    if isinstance(cls, str):
        cls = importer(cls)
    expender = cls(**kwargs)
    expender.fix_abs_path()


def fix_retry_env(**kwargs):
    """
    This method "fixes" os.environ making path to Work both on Windows and Linux.

    For each key in ENV_KEYS, this method check if the value can be successfully resolved
    as path. If so, it does nothing. Otherwise, it strip away the drive part (i.e. C:\\)
    changing os.environ entry.


    :param ENV_KEYS keys of os.environ which will be "fixed".
    :param ENV_KEY_SEP: seperator used in ENV_PCK. Optional. Default is ','
    :param ENV_SEP: seperator that is used inside path. Optional. Default is os.path.sep.
    :param ENV_DELIM_SEP: seperator that is used inside os.environ. Optional. Default is os.pathsep.
    :param cls: class or str with implementation logic. Optional. Default is OsEnvrionPathRetry.
    """

    cls = kwargs.pop('cls', OsEnvrionPathRetry)
    if isinstance(cls, str):
        cls = importer(cls)
    path_retry = cls(**kwargs)
    path_retry.fix_retry_path()


import logging
import sys
import pytest

logger = logging.getLogger(__name__)

try:
    #python3 -m pip uninstall multidispatch==0.2
    from multidispatch import multimethod
except ImportError:
    class _dummy(object):
        def dispatch(self, *types):
            return self

    def multimethod(func):
        def with_md(*args, **kwargs):
            return _dummy
        return with_md

from alexber.utils.importer import importer


class Engine(object):


    @multimethod(object)
    def _set_playerA(self, player):
        logging.debug('instance')
        pass

    @_set_playerA.dispatch(str)
    def _(self, st):
        logging.debug('str')
        player_cls = importer(st)
        player = player_cls()
        player.say()
        pass

    #alternative (recent) syntax
    # @_set_playerA.dispatch
    # def _(self, st: str):
    #     print('str')
    #     player_cls = importer(st)
    #     player = player_cls()
    #     player.say()
    #     pass


    playerA = property(fset=_set_playerA)


class Player(object):
    def say(self):
        logging.debug('hello')

@pytest.mark.md
def test_overloading_object(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    namespace = sys.modules[__name__]

    importer = mocker.spy(namespace, 'importer')
    engine = Engine()
    input = Player()
    engine.playerA= input

    assert namespace.importer == importer
    assert importer.call_count ==0, "Engine.playerA(str) was called, when Engine.playerA(object) was expected"
                                                            # #we shouldn't use  importer when we have explicit object."
    logger.info(dir(importer))

@pytest.mark.md
def test_overloading_str(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    namespace = sys.modules[__name__]

    importer = mocker.spy(namespace, 'importer')
    engine = Engine()
    input = '.'.join([__name__, 'Player'])
    engine.playerA = input

    assert namespace.importer == importer
    importer.assert_called_once_with(input)

class PlayerPhilosopher:
    def __init_subclass__(cls, default_name, **kwargs):
        super().__init_subclass__(**kwargs)
        logger.debug(f"Called __init_subclass({cls}, {default_name})")
        cls.default_name = default_name


class PlayerAustralianPhilosopher(PlayerPhilosopher, default_name="Bruce"):
    pass



if __name__ == "__main__":
    import pytest
    pytest.main([__file__])


#one-time setup
import os
cwd = os.getcwd()
if cwd.endswith('AlexBerUtils'):
    os.chdir(os.path.join(cwd, 'data'))

import pytest

ismultidispatchfound = False
try:
    # pip install multidispatch==0.2
    #python3 -m pip install .[md]
    from multidispatch import multimethod
    ismultidispatchfound = True
except ImportError:
    pass

isyamlrelatedfound = False
try:
    #python3 -m pip install .[yaml]
    import alexber.utils._ymlparsers_extra as ymlparsers_extra
    if ymlparsers_extra._isHiYaPyCoAvailable and ymlparsers_extra._isJinja2DefaultAvailable:
        isyamlrelatedfound = True
except ImportError:
    pass


#see https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
def skip_tests(config=None, items=None, keyword=None, reason=None):
    if items is None:
        TypeError("items can't be None")

    if reason is None:
        TypeError("reason can't be None")

    if keyword is None:
        TypeError("keyword can't be None")

    skip = pytest.mark.skip(reason=reason)
    for item in items:
        if keyword in item.keywords:
            item.add_marker(skip)



#see https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
def pytest_collection_modifyitems(config, items):
    if not ismultidispatchfound:
        skip_tests(items=items, keyword="md", reason="multidispatch is not installed..")

    if not isyamlrelatedfound:
        skip_tests(items=items, keyword="yml", reason="yml is not installed..")

#see https://docs.pytest.org/en/latest/mark.html
#see https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
def pytest_configure(config):
    config.addinivalue_line(
        "markers", "yml: yml is not installed.."
    )
    config.addinivalue_line(
        "markers", "md: multidispatch is not installed.."
    )


import logging
from collections import OrderedDict
from collections.abc import Mapping
import pytest
import copy

logger = logging.getLogger(__name__)
from alexber.utils.init_app_conf import mask_value, merge_list_value_in_dicts, to_convex_map, \
    parse_config
import alexber.utils.init_app_conf as init_app_conf
from alexber.utils.init_app_conf import _create_default_parser, AppConfParser
from alexber.utils.parsers import is_empty
from tests.utils.ymlparsers_test import ymlparsersSetup, ymlparsersCleanup, exp_config_d
from importlib.resources import path

def create_default_parser(implicit_convert=None):
    parser=_create_default_parser(**{'implicit_convert': implicit_convert})
    return parser

def _reset_initappconf():
    init_app_conf.default_parser_cls  = None
    init_app_conf.default_parser_kwargs = None



@pytest.fixture
def initappconfFixture(mocker):
    _reset_initappconf()

    init_app_conf.initConfig()
    yield None
    _reset_initappconf()

@pytest.fixture
def initappconfFalseFixture(mocker):
    _reset_initappconf()

    init_app_conf.initConfig(**{'default_parser_kwargs': {'implicit_convert': False}})
    yield None
    _reset_initappconf()



class MyAppConfParser(AppConfParser):
    pass

def test_mask_value_default_as_none(request, initappconfFalseFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    with pytest.raises(ValueError, match='None'):
        init_app_conf.initConfig(**{'default_parser_kwargs': {'implicit_convert': None}})
        mask_value('0.1')

def test_initConfig_default_parser_cls_str(request, initappconfFalseFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    class_name = '.'.join([__name__, MyAppConfParser.__name__])
    init_app_conf.initConfig(**{'default_parser_cls': class_name})

    pytest.assume(init_app_conf.default_parser_cls == MyAppConfParser)



@pytest.mark.parametrize(
    'value, exp_value, exp_type',

    [
        ('John', 'John', str),
        ('1000', '1000', str),
        ('0.1', '0.1', str),
        ('None', 'None', str),
        ('True', 'True', str),
        ('False', 'False', str),


    ]
)
def test_mask_value_without_implicit_convert(request, initappconfFixture, value, exp_value, exp_type):
    logger.info(f'{request._pyfuncitem.name}()')


    result = mask_value(value, implicit_convert=False)
    type_result = type(result)
    pytest.assume(exp_value == result)
    pytest.assume(exp_type == type_result)


@pytest.mark.parametrize(
     'value, exp_value, exp_type',

    [
        ('John', 'John', str),
        ('alexber.utils.players.ConstantPlayer', 'alexber.utils.players.ConstantPlayer', str),
        ('1000', 1000, int),
        ('0.1', 0.1, float),
        ('0.0', 0.0, float),
        ('-0.0', 0.0, float),
        ('-5', -5, int),
        ('None', None, type(None)),
        ('True', True, bool),
        ('False', False, bool),



    ]
)
def test_mask_value(request, initappconfFixture, value, exp_value, exp_type):
    logger.info(f'{request._pyfuncitem.name}()')


    result = mask_value(value)
    type_result = type(result)
    pytest.assume(exp_value == result)
    pytest.assume(exp_type == type_result)

def test_mask_value_default_as_true(request, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    value = '0.1'
    exp_value = 0.1
    exp_type = float

    result = mask_value(value)  #implicit_convert=None as True
    type_result = type(result)
    pytest.assume(exp_value == result)
    pytest.assume(exp_type == type_result)

def test_mask_value_default_as_false(request, initappconfFalseFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    value = '0.1'
    exp_value = '0.1'
    exp_type = str

    result = mask_value(value)  #implicit_convert=None as False
    type_result = type(result)
    pytest.assume(exp_value == result)
    pytest.assume(exp_type == type_result)




def _calc_exp_ports(ports):
    if ports is None:
        return None
    arr = str(ports).split(',')
    ret=[mask_value(port) for port in arr]
    return ret


@pytest.mark.parametrize(
     'value, exp_value',

    [
         (10000, [10000]),
        ('10000', [10000]),
        ('10000,', [10000, None]),
        ('10000,10001',  [10000,10001]),
        ('10000,10001,', [10000,10001, None]),
        ('10000,,10003', [10000,None, 10003]),

        (True, [True]),
        (False, [False]),
        ('True', [True]),
        ('False', [False]),
        ('False,True,False', [False,True,False]),

        ([False], [False]), #Not_Supported
        ([None], [None]), #Not_Supported
        (['False'], [False]), #Not_Supported
        (['None'], ['None']), #Not_Supported

        ([None], ["innervalue"]),
        ([''], ["innervalue"]),
    ]
)
def test_merge_list_value_in_dicts(request, initappconfFixture, value, exp_value):
    logger.info(f'{request._pyfuncitem.name}()')
    if type(value) is list:
        logger.debug(f"Complex types (such as list) are not supported.")
        return

    flat_d = OrderedDict()
    flat_d['app.ports'] = value
    d = OrderedDict()
    inner_value = ["innervalue"]
    d.setdefault('app', OrderedDict()).setdefault('ports', inner_value)

    ret = merge_list_value_in_dicts(flat_d, d, 'app', 'ports', implicit_convert=True)
    f_value = flat_d['app.ports']
    i_value = d['app']['ports']

    pytest.assume(value==f_value)
    pytest.assume(inner_value==i_value)
    pytest.assume(exp_value==ret)

def test_merge_list_value_in_dicts_absent(request, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    flat_d = OrderedDict()
    d = OrderedDict()
    ret = merge_list_value_in_dicts(flat_d, d, 'general', 'profiles')
    b = is_empty(ret)
    pytest.assume(b)

def test_merge_list_value_in_dicts_default_as_true(request, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    flat_d = {'app.ports': 10000}
    d = {'app': {'ports': ["innervalue"]}}
    inner_value = ["innervalue"]

    ret = merge_list_value_in_dicts(flat_d, d, 'app', 'ports')
    f_value = flat_d['app.ports']
    i_value = d['app']['ports']

    pytest.assume(10000==f_value)
    pytest.assume(inner_value==i_value)
    pytest.assume([10000]==ret)


def test_merge_list_value_in_dicts_default_as_false(request, initappconfFalseFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    flat_d = {'app.ports': 10000}
    d = {'app': {'ports': ["innervalue"]}}
    inner_value = ["innervalue"]

    ret = merge_list_value_in_dicts(flat_d, d, 'app', 'ports')
    f_value = flat_d['app.ports']
    i_value = d['app']['ports']

    pytest.assume(10000==f_value)
    pytest.assume(inner_value==i_value)
    pytest.assume(['10000']==ret)

def test_merge_list_value_in_dicts_false(request, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    flat_d = {'app.ports': 10000}
    d = {'app': {'ports': ["innervalue"]}}
    inner_value = ["innervalue"]

    ret = merge_list_value_in_dicts(flat_d, d, 'app', 'ports', implicit_convert=False)
    f_value = flat_d['app.ports']
    i_value = d['app']['ports']

    pytest.assume(10000==f_value)
    pytest.assume(inner_value==i_value)
    pytest.assume(['10000']==ret)





def check_exp_sys(exp_d, default_d):
    general_d = default_d.get('general', None)
    assert general_d is not None
    app_d = default_d.get('app', None)
    assert app_d is not None
    app_str_d = app_d.get('as_str', None)
    assert app_str_d is not None

    exp_val = exp_d['general.profiles']
    val = general_d.get('profiles', None)
    pytest.assume(exp_val==val)

    for flat_key, exp_val in exp_d.items():
        if flat_key.startswith("general"):
            continue
        arr = flat_key.split('.')
        key = arr[-1]
        dd = app_str_d if 'as_str' in flat_key else app_d
        val = dd.get(key, None)
        pytest.assume(exp_val == val)


def test_to_convex_map_intented(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {'general.profiles': ["dev"],
             'app.white_list': ['one', 'two', 'three'],
             'app.as_str.alt_white_list': [10],
             'app.num_scalar': 5,
             'app.none_scalar': None,
             'app.str_scalar': 'Hello',
             'app.as_str.bool1_scalar': True,
             'app.as_str.bool2_scalar': False,

             'app.as_str.list_bool': [True, False],
             'app.as_str.list_bool_true': [True],
             'app.as_str.list_bool_false': [False],
             'app.as_str.list_none': [None],
             'app.as_str.list_empty_str': [''],
             }

    sys_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.num_scalar': 5,
             'app.none_scalar': None,
             'app.str_scalar': 'Hello',
             'app.as_str.bool1_scalar': 'True',
             'app.as_str.bool2_scalar': 'False',

             'app.as_str.list_bool': 'True,False',
             'app.as_str.list_bool_true': True,
             'app.as_str.list_bool_false': False,
             'app.as_str.list_none': None,
             'app.as_str.list_empty_str': '',
             }
    list_ensure = [key for key in sys_d.keys() if 'scalar' not in key]
    create_default_parser()._do_ensure_list(sys_d, list_ensure)
    default_d = to_convex_map(sys_d)

    check_exp_sys(exp_d, default_d)



def test_to_convex_map_whitelist(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {'general.profiles': ["dev"],
             'app.as_str.alt_white_list': [10],
             'app.as_str.bool1_scalar': True,
             }

    sys_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             'db.unused': 'localhost',
             }
    list_ensure = [key for key in sys_d.keys() if 'scalar' not in key]
    create_default_parser()._do_ensure_list(sys_d, list_ensure)
    default_d = to_convex_map(sys_d, white_list_flat_keys=['general', 'app.as_str'], implicit_convert=True)

    check_exp_sys(exp_d, default_d)

def test_to_convex_map_whitelist_implicit_convert_false(request, mocker, initappconfFalseFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {'general.profiles': ["dev"],
             'app.as_str.alt_white_list': ['10'],
             'app.as_str.bool1_scalar': 'True',
             }

    sys_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             'db.unused': 'localhost',
             }
    list_ensure = [key for key in sys_d.keys() if 'scalar' not in key]
    create_default_parser()._do_ensure_list(sys_d, list_ensure)
    default_d = to_convex_map(sys_d, white_list_flat_keys=['general', 'app.as_str'], implicit_convert=False)

    check_exp_sys(exp_d, default_d)

def test_to_convex_map_whitelist_implicit_convert_false_default_as_false(request, mocker, initappconfFalseFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {'general.profiles': ["dev"],
             'app.as_str.alt_white_list': ['10'],
             'app.as_str.bool1_scalar': 'True',
             }

    sys_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             'db.unused': 'localhost',
             }
    list_ensure = [key for key in sys_d.keys() if 'scalar' not in key]
    create_default_parser()._do_ensure_list(sys_d, list_ensure)
    default_d = to_convex_map(sys_d, white_list_flat_keys=['general', 'app.as_str'], implicit_convert=False)

    check_exp_sys(exp_d, default_d)

def test_to_convex_map_whitelist_default_as_true(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {'general.profiles': ["dev"],
             'app.as_str.alt_white_list': [10],
             'app.as_str.bool1_scalar': True,
             }

    sys_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             'db.unused': 'localhost',
             }
    list_ensure = [key for key in sys_d.keys() if 'scalar' not in key]
    create_default_parser()._do_ensure_list(sys_d, list_ensure)
    default_d = to_convex_map(sys_d, white_list_flat_keys=['general', 'app.as_str'])

    check_exp_sys(exp_d, default_d)

@pytest.mark.parametrize(
     'exp_profiles, sys_d, default_d',
    [
        (['dev', 'local'], {'general.profiles': "dev,local"}, {'general': {'profiles': ["dev"]}}),
        (['prod'], {'general.profiles': "prod"}, {'general': {'profiles': ["dev"]}}),
        (['dev'], {}, {'general': {'profiles': ["dev"]}}),
        (['prod'], {'general.profiles': "prod"}, {}),
        (['prod'], {'general.profiles': "prod"}, {'general': {'profiles': None}}),
        (['prod'], {'general.profiles': "prod"}, {'general': {'profiles': [""]}}),
        ([None], {'general.profiles': None}, {'general': {'profiles': ["dev"]}}),
        ([""], {'general.profiles': ""}, {'general': {'profiles': ["dev"]}}),
        ([], {'general.log': "something"}, {'general': {'log': {'root': 'INFO'}}}), #it should be list, not dict
    ]
)
def test_parse_profiles(request, mocker, initappconfFixture, exp_profiles, sys_d, default_d):
    logger.info(f'{request._pyfuncitem.name}()')

    profiles = create_default_parser()._parse_profiles(sys_d, default_d)
    pytest.assume(exp_profiles == profiles)


@pytest.mark.parametrize(
     'exp_value, default_d',

    [
        (False, {'general': {'whiteListSysOverride': ["general,app"]}}),
        (False, {'general': {'whiteListSysOverride': ["general"]}}),
        (True, {'general': {'whiteListSysOverride': []}}),
        (False, {'general': {'whiteListSysOverride': ['']}}),
        (False, {'general': {'whiteListSysOverride': [None]}}),
        (True, {'general': {'whiteListSysOverride': None}}),
    ]

)

def test_parse_white_list_implicitely(request, mocker, initappconfFixture, exp_value, default_d):
    logger.info(f'{request._pyfuncitem.name}()')
    result = create_default_parser()._parse_white_list_implicitely(default_d)
    pytest.assume(exp_value==result)




@pytest.mark.parametrize(
     'default_d',

    [
        {'general': {'profiles': ["dev"]}, 'some_unique_key': 'value'},
        {'general': {'profiles': ["dev"], 'whiteListSysOverride':None}, 'some_unique_key': 'value'},
        {'general': {'profiles': ["dev"], 'whiteListSysOverride':[]}, 'some_unique_key': 'value'},
    ]

)

def test_parse_white_list_implicit(request, mocker, initappconfFixture, default_d):
    logger.info(f'{request._pyfuncitem.name}()')
    #whiteListSysOverride

    white_list = create_default_parser()._parse_white_list(default_d)
    length = 0 if white_list is None else len(white_list)
    pytest.assume(length==2)
    pytest.assume('general' in white_list)
    pytest.assume('some_unique_key' in white_list)

def test_parse_white_list_explicit(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    #whiteListSysOverride
    default_d={'general': {'profiles': ["dev"], 'whiteListSysOverride': ['general']},
               'app': {
                   'ignoredKey': 1,
               }}

    white_list = create_default_parser()._parse_white_list(default_d)
    length = 0 if white_list is None else len(white_list)
    pytest.assume(length==1)
    pytest.assume('general' in white_list)
    pytest.assume('app' not in white_list)

def test_parse_white_list_explicit2(request, initappconfFixture, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    #whiteListSysOverride

    default_d={'general': {'profiles': ["dev"],
                           'log': None,
                           'whiteListSysOverride': ['general', 'app']},
               'app': {
                  'ignoredKey': 1,
                }
    }

    white_list = create_default_parser()._parse_white_list(default_d)
    length = 0 if white_list is None else len(white_list)
    pytest.assume(length==2)
    pytest.assume('general' in white_list)
    pytest.assume('app' in white_list)


def test_parse_list_ensure(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    sys_d = {}
    default_d={'general': {'listEnsure': ['general.profiles', 'general.listEnsure',
                                          'general.listEnsure', 'general.whiteListSysOverride',
                                          'app.check_list']}}

    list_ensure = create_default_parser()._parse_list_ensure(sys_d, default_d)
    length = 0 if list_ensure is None else len(list_ensure)
    pytest.assume(length==1)
    pytest.assume('app.check_list' in list_ensure)


def test_parse_list_ensure_empty(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    sys_d = {}
    default_d={'general': {'listEnsure': []}}

    list_ensure = create_default_parser()._parse_list_ensure(sys_d, default_d)
    length = 0 if list_ensure is None else len(list_ensure)
    pytest.assume(length==0)

def test_parse_list_ensure2(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    sys_d = {'general.listEnsure':'general.profiles,app.white_list'}

    default_d={'general': {'listEnsure': ['general.profiles', 'general.listEnsure',
                                          'general.listEnsure', 'general.whiteListSysOverride',
                                          'app.check_list']}}

    list_ensure = create_default_parser()._parse_list_ensure(sys_d, default_d)
    length = 0 if list_ensure is None else len(list_ensure)
    pytest.assume(length==1)
    pytest.assume('app.white_list' in list_ensure)

def test_get_white_listed_empty_src_d(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {}
    d = create_default_parser()._get_white_listed({}, {'a':'b'})
    pytest.assume(exp_d==d)

def test_get_white_listed_empty_white_list_flat_keys(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             'db.unused': 'localhost',
             }

    sys_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             'db.unused': 'localhost',
             }

    d = create_default_parser()._get_white_listed(sys_d, None)
    pytest.assume(exp_d==d)

def test_get_white_listed(request, mocker, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    exp_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             }

    sys_d = {'general.profiles': "dev",
             'app.white_list': 'one,two,three',
             'app.as_str.alt_white_list': '10',
             'app.as_str.bool1_scalar': 'True',
             'db.unused': 'localhost',
             }

    white_list_flat_keys = ['app', 'general']

    d = create_default_parser()._get_white_listed(sys_d, white_list_flat_keys)
    pytest.assume(exp_d==d)

@pytest.mark.yml
def test_parse_sys_args(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    expdd = {
        'general': {'profiles': ['dev'],  # list
                    'log': {
                        'formatters': {
                            'detail': {
                                'format': '%(message)s'
                            }
                        },
                        'root': {
                            'level': 20  # logging.INFO
                        }
                    },
                    'listEnsure': ['app.white_list','app.alt_white_list']
                    },  # list
        'app': {'host_name': 'yahoo.com',
                'portal': 'reddit.com',     #this will be proceses because of implicit general.whiteListSysOverride
                'white_list': ['alpha', 'betha', 'gamma'],
                'alt_white_list': [],  # not part of sys.args, so value from config.yml will be ignored
                }
    }

    pck = '.'.join(['tests_data', __package__, 'initappconf'])
    with path(pck, 'config.yml') as full_path:
        argsv = f'--general.config.file={full_path} '\
                '--general.profiles=dev ' \
                '--general.log.formatters.detail.format=%(message)s ' \
                '--general.log.root.level=20 ' \
                '--general.listEnsure=app.white_list,app.alt_white_list ' \
                '--app.host_name=yahoo.com ' \
                '--app.portal=reddit.com ' \
                '--app.white_list=alpha,betha,gamma ' \
            .split()


        dd, profiles, _, list_ensure, _ = create_default_parser()._parse_sys_args(None, argsv)
        dd['general']['profiles'] = profiles
        dd['general']['listEnsure'] = list_ensure

        pytest.assume(expdd==dd)

@pytest.mark.yml
def test_parse_yml(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    expdd = {
        'general': {'profiles': ['dev'],  # list
                    'log': {
                        'version': 1,
                        'disable_existing_loggers': False,
                        'formatters': {
                            'brief': {
                                'format': '%(message)s'
                            },
                            'detail': {
                                'format': '%(asctime)-15s %(levelname)s [%(name)s.%(funcName)s] %(message)s',
                                'datefmt': '%Y-%m-%d %H:%M:%S'
                            },
                        },
                        'root': {
                            'level': 'INFO'  # logging.INFO
                        }
                    },
                    'listEnsure': ['app.white_list', 'app.alt_white_list']
                    },  # list
        'app': {'host_name': 'yahoo.com',
                'portal': 'reddit.com',  # this will be proceses because of implicit general.whiteListSysOverride
                'news': 'cnn.com',
                'white_list': ['alpha', 'betha', 'gamma'],
                'alt_white_list': [100, 10.0],  #value from config.yml will be parsed
                }
    }



    sys_d = {
        'general': {'profiles': ['dev'],
                    'listEnsure': ['app.white_list','app.alt_white_list']
                    },
        'app': {'host_name': 'yahoo.com',
                'portal': 'reddit.com',     #this will be proceses because of implicit general.whiteListSysOverride
                'white_list': ['alpha', 'betha', 'gamma'],
                }
    }

    pck = '.'.join(['tests_data', __package__, 'initappconf'])
    with path(pck, 'config.yml') as full_path:
        dd = create_default_parser()._parse_yml(sys_d, ['dev'], config_file=full_path)
        del dd['general']['log']['handlers']
        del dd['general']['log']['root']['handlers']
        pytest.assume(expdd==dd)

@pytest.mark.yml
def test_parse_config_true(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    expdd = {
        'general': {'profiles': ['dev'],
                    'log': {
                        'version': 1,
                        'disable_existing_loggers': False,
                        'formatters': {
                            'brief': {
                                'format': '%(message)s'
                            },
                            'detail': {
                                'format': '%(asctime)-15s %(levelname)s [%(name)s.%(funcName)s] %(message)s',
                                'datefmt': '%Y-%m-%d %H:%M:%S'
                            },
                        },
                        'root': {
                            'level': 'INFO'  # logging.INFO
                        }
                    },
                    'config':{
                        'file' : None   #will be be populated below
                    },
                    'whiteListSysOverride': ['general', 'app'],   #implictely populated
                    'listEnsure': [],   #profile can be overridden in sysargs, regargless,  \
                                        # we don't override any (other) list value from sysargs
                    },  # list
        'app': {'host_name': 'yahoo.com',
                'news': 'cnn.com',
                'portal': 'reddit.com',     #this will be proceses because of implicit general.whiteListSysOverride
                'white_list': ['gamma', 'alpha', 'betha'],
                'alt_white_list': [100, 10.0],
                }
    }

    pck = '.'.join(['tests_data', __package__, 'initappconf'])

    with path(pck, 'config.yml') as full_path:
        expdd['general']['config']['file'] = str(full_path)

        argsv = f'--general.config.file={full_path} '\
                '--app.portal=reddit.com ' \
                '--general.profiles=dev ' \
            .split()


        dd = parse_config(None, argsv)
        del dd['general']['log']['handlers']
        del dd['general']['log']['root']['handlers']

        pytest.assume(expdd==dd)

@pytest.mark.yml
def test_parse_config_explicit_white_list(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')

    expdd = {
        'general': {'profiles': [],
                    'log': {
                        'version': 1,
                        'disable_existing_loggers': False,
                        'formatters': {
                            'brief': {
                                'format': '%(message)s'
                            },
                            'detail': {
                                'format': '%(asctime)-15s %(levelname)s [%(name)s.%(funcName)s] %(message)s',
                                'datefmt': '%Y-%m-%d %H:%M:%S'
                            },
                        },
                        'root': {
                            'level': 'DEBUG'
                        }
                    },
                    'config':{
                        'file' : None   #will be be populated below
                    },
                    'whiteListSysOverride': ['app.host_name'],   #explicetly populated
                    'listEnsure': [],   #whiteListSysOverride can be overridden in sysargs, regargless,  \
                                        # we don't override any (other) list value from sysargs
                    },  # list
        'app': {'host_name': '10.20.40.60',
                'news': 'cnn.com',
                #'portal': 'reddit.com',     #this will NOT be proceses because of explicit general.whiteListSysOverride
                'white_list': ['gamma', 'alpha', 'betha'],
                'alt_white_list': [100, 10.0],
                }
    }

    pck = '.'.join(['tests_data', __package__, 'initappconf'])

    with path(pck, 'confige.yml') as full_path:
        expdd['general']['config']['file'] = str(full_path)

        #app.host_name is in general.whiteListSysOverridem it will be process
        #app.portal and last line will be ignored
        argsv = f'--general.config.file={full_path} '\
                '--app.portal=reddit.com ' \
                '--app.host_name=10.20.40.60 ' \
                '10.20.40.60:8082 ' \
            .split()


        dd = parse_config(None, argsv)
        del dd['general']['log']['handlers']
        del dd['general']['log']['root']['handlers']

        pytest.assume(expdd==dd)

@pytest.mark.yml
def test_parse_config_false(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    pck = '.'.join(['tests_data', __package__, 'initappconf'])
    exp_disable_existing_loggers = 'False'

    with path(pck, 'config.yml') as full_path:
        argsv = f'--general.config.file={full_path} ' \
                '--general.profiles=dev ' \
                f'--general.log.disable_existing_loggers={exp_disable_existing_loggers} ' \
            .split()

        dd = parse_config(None, argsv, implicit_convert=False)
        disable_existing_loggers = dd['general']['log']['disable_existing_loggers']
        pytest.assume(exp_disable_existing_loggers==disable_existing_loggers)

@pytest.mark.yml
def test_parse_config_default_as_false(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFalseFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    pck = '.'.join(['tests_data', __package__, 'initappconf'])
    exp_disable_existing_loggers = 'False'

    with path(pck, 'config.yml') as full_path:
        argsv = f'--general.config.file={full_path} ' \
                '--general.profiles=dev ' \
                f'--general.log.disable_existing_loggers={exp_disable_existing_loggers} ' \
            .split()

        dd = parse_config(None, argsv)
        disable_existing_loggers = dd['general']['log']['disable_existing_loggers']
        pytest.assume(exp_disable_existing_loggers==disable_existing_loggers)

@pytest.mark.yml
def test_parse_config_default_as_true(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFixture):
    logger.info(f'{request._pyfuncitem.name}()')
    pck = '.'.join(['tests_data', __package__, 'initappconf'])
    exp_disable_existing_loggers = False

    with path(pck, 'config.yml') as full_path:
        argsv = f'--general.config.file={full_path} ' \
                '--general.profiles=dev ' \
                f'--general.log.disable_existing_loggers={exp_disable_existing_loggers} ' \
            .split()

        dd = parse_config(None, argsv)
        disable_existing_loggers = dd['general']['log']['disable_existing_loggers']
        pytest.assume(exp_disable_existing_loggers==disable_existing_loggers)



if __name__ == "__main__":
    pytest.main([__file__])





import logging
import os as _os
import subprocess
import threading
from concurrent.futures import ThreadPoolExecutor
from . importer import importer

executor = None
default_log_name  = None
default_log_level = None
default_logpipe_cls = None
default_log_subprocess_cls = None

#inspired by https://codereview.stackexchange.com/questions/6567/redirecting-subprocesses-output-stdout-and-stderr-to-the-logging-module/175382
#for alternatives see https://gist.github.com/jaketame/3ed43d1c52e9abccd742b1792c449079
# that is itself adoptation of https://gist.github.com/bgreenlee/1402841


class BasePipe(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fdRead, self.fdWrite = _os.pipe()
        self.fdWriteLock = threading.RLock()

        with self.fdWriteLock:
            self.is_closed = False

    def fileno(self):
        """
        Return the write file descriptor of the pipe
        """
        return self.fdWrite

    def processLine(self, line):
        """
        Hook to be overridden in the sub-class.
        """
        pass

    def run(self):
        """
        Run the thread, process output line by line.
        """
        with _os.fdopen(self.fdRead) as pipeReader:
            for line in iter(pipeReader.readline, ''):
                self.processLine(line)

    def breakPipe(self):
        self.close()

    def cleanUp(self):
        """
        Hook to be overridden in the sub-class.
        """
        pass

    def close(self):
        """
        Close the write end of the pipe.
        """
        with self.fdWriteLock:
            if not self.is_closed:
                self.cleanUp()
                _os.close(self.fdWrite)
            self.is_closed = True

    def __enter__(self):
        if self.is_closed:
            raise ValueError("I/O operation on closed pipe")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class LogPipe(BasePipe):
    def __init__(self, *args, **kwargs):
        """
        Setup the logger with logName and loglevel
        Create read and write file descriptor of the pipe
        """
        logName = kwargs.pop('logName', default_log_name)
        logLevel = kwargs.pop('logLevel', default_log_level)

        super().__init__(*args, **kwargs)

        self.logger = logging.getLogger(logName)

        self.level = logLevel

    def processLine(self, line):
        self.logger.log(self.level, line.rstrip('\r\n'))

class FilePipe(BasePipe):
    def __init__(self, *args, **kwargs):
        filename = kwargs.pop('fileName', None)
        if filename is None:
            raise ValueError('fileName should be specified')

        super().__init__(*args, **kwargs)

        d = {
            'file': filename,
            'mode': 'w',
            **kwargs
        }

        self.f = open(**d)

    def processLine(self, line):
        aline = f"{line.rstrip()}\n"
        self.f.write(aline)

    def cleanUp(self):
        if self.f is not None:
            self.f.close()





class LogSubProcessCall(object):
    def __init__(self, *args, **kwargs):

        self.pipe = kwargs.pop('pipe')
        d = kwargs.pop('popen', {})

        self.popenargs = d.pop('args', [])
        self.popenkwargs = d.pop('kwargs', {})
        super().__init__(*args, **kwargs)

    def calc_subprocess_run_kwargs(self):
        kwargs= {'stdout':self.pipe, 'stderr':subprocess.STDOUT,
            'text':True, 'bufsize':1, 'check':True,
                 **self.popenkwargs}
        return kwargs

    def run_sub_process(self):
        f = executor.submit(self.pipe.run)
        try:
            kwargs = self.calc_subprocess_run_kwargs()
            process = subprocess.run(self.popenargs, **kwargs)
        finally:
            self.pipe.breakPipe()
            f.result()



def run_sub_process(*args, **kwargs):
    """
    This method run subprocess and logs it's output to the logger.
    This method is sophisticated decorator to subprocess.run(). It is useful, when your subprocess
    run's a lot of time and you're interesting to receive it's stdout and stderr. By default, it's streamed to log.
    You can easily customize this behavior, see `initConig()` method.

    This method is sophisticated decorator to subprocess.run(). See it's docstring for more information. Note, that
    some parameters (cwd, for example) that can be used in popenkwargs are listed in Popen constructor.

    logPipe is customizable object that essentially forwards output from subprocess to the logger using
    logName and logLevel (see below).

    default_log_subprocess_cls by default is LogSubProcessCall.
    default_logpipe_cls by default is LogPipe.
    logName by default is processinvokes.
    logLevel by default is logger.INFO.
    See initConfig() for more details.

    Default parameters for subprocess.run() are (it can be overridden in popenkwargs):
        'stdout':logPipe,
        'stderr':STDOUT,
        'text':True,
        'bufsize': 1,
        'check': True

    This means:
        logPipe is used as subprocess's standard output and standard error.
        Do decode stdin, stdout and stderr using the given encoding
          or the system default otherwise.
        1 is supplied as the buffering argument to the open() function when
          creating the stdin/stdout/stderr pipe file objects.
          Essentially, no OS-level buffering between process, provided that call to write contains a newline character.
        If the exit code of subprocess is non-zero, it raises a CalledProcessError.

    It is generally not advice to override them, but you can if you know, what you're doing.

    :param args: will be passed as popenargs to subprocess.run() method
    :param roughly kwargs['kwargs'] will be passed as popenkwargs to subprocess.run() method
    :param roughly kwargs['logPipe']['cls'] or default_logpipe_cls (if first one is empty)
                                             will be passed as logPipeCls to create logPipe.
                                             Can be class or str.
    :param roughly kwargs['logPipe']['kwargs'] will be passed as kwargs to logPipeCls to create logPipe.
    :param roughly kwargs['logSubprocess']['cls'] or default_log_subprocess_cls (if first one is empty)
                                             will be passed as logSubProcessCls to create LogSubProcessCall.
                                              Can be class or str.
    :param roughly kwargs['logSubprocess']['kwargs'] will be passed as kwargs to logSubprocess.

    :return:
    """
    if default_logpipe_cls is None:
        raise ValueError("default_logpipe_cls can't be None")

    logPipe_p = kwargs.pop('logPipe', {})
    logPipeCls = logPipe_p.pop('cls', default_logpipe_cls)
    if isinstance(logPipeCls, str):
        logPipeCls = importer(logPipeCls)

    logPipeKwargs = logPipe_p.pop('kwargs', {})
    callKwargs = kwargs.pop('kwargs', {})
    logSubprocess_p = kwargs.pop('logSubprocess', {})
    logSubProcessCls = logSubprocess_p.pop('cls', default_log_subprocess_cls)
    if isinstance(logSubProcessCls, str):
        logSubProcessCls = importer(logSubProcessCls)
    logSubprocessKwargs = logSubprocess_p.pop('kwargs', {})

    with logPipeCls(**logPipeKwargs) as logPipe:

        kwargs = {'pipe': logPipe,
                 'popen':
                      {'args': args,
                       'kwargs': callKwargs
                      },
                 **logSubprocessKwargs
                  }

        call = logSubProcessCls(**kwargs)
        call.run_sub_process()



def initConfig(**kwargs):
    """
    This method can be optionally called prior any call to another function in this module.
    It is indented to be called in the MainThread.
    This method can be call with empty params.

    :param default_log_name: Optional. - name of the logger where the messages will be streamed to.
                Default values is: processinvokes
    :param default_log_level: Optional. - log level to be used in logger.
                Default values is: logging.INFO
    :param default_logpipe_cls: can be class or str. Optional. You can use your custom class for the logging.
                For example, FilePipe.
                Default values is: LogPipe
    :param default_log_subprocess_cls: can be class or str. Optional.
                Default values is: LogSubProcessCall
    :param executor: internally used to run sub-process
                Default values are
                      'max_workers':1,
                      'thread_name_prefix': processinvokes
                This means, by default:
                    We're using up to 1 worker.
                    In log message generated from the worker processinvokes-xxx will be used as thread_name.

    If running from the MainThread, this method is idempotent.
    :return:
    """
    default_log_name_p = kwargs.get('default_log_name', None)
    if default_log_name_p is None:
        default_log_name_p = __name__
    global default_log_name
    default_log_name = default_log_name_p

    default_log_level_p = kwargs.get('default_log_level', None)
    if default_log_level_p is None:
        default_log_level_p = logging.INFO
    global default_log_level
    default_log_level = default_log_level_p

    default_logpipe_cls_p = kwargs.get('default_logpipe_cls', None)
    if default_logpipe_cls_p is None:
        default_logpipe_cls_p = LogPipe
    elif isinstance(default_logpipe_cls_p, str):
        default_logpipe_cls_p = importer(default_logpipe_cls_p)
    global default_logpipe_cls
    default_logpipe_cls = default_logpipe_cls_p

    default_log_subprocess_cls_p = kwargs.get('default_log_subprocess_cls', None)
    if default_log_subprocess_cls_p is None:
        default_log_subprocess_cls_p = LogSubProcessCall
    elif isinstance(default_log_subprocess_cls_p, str):
        default_log_subprocess_cls_p = importer(default_log_subprocess_cls_p)
    global default_log_subprocess_cls
    default_log_subprocess_cls = default_log_subprocess_cls_p

    executor_d = kwargs.get('executor', {})
    executor_d = {'max_workers': 1,
                  'thread_name_prefix': __name__,
                  **executor_d}
    global executor
    if executor is not None:
        executor.shutdown(wait=False)
    executor = ThreadPoolExecutor(**executor_d)

initConfig()


import logging
import pytest
import threading

logger = logging.getLogger(__name__)

from alexber.utils.thread_locals import threadlocal_var, get_threadlocal_var, del_threadlocal_var
import random

def test_get_threadlocal_var_empty(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


def test_get_threadlocal_var_exist(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    expValue = 1
    thread_locals.value = expValue

    value = get_threadlocal_var(thread_locals, 'value')
    assert expValue == value

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


ns = threading.local()
stop = 10

class Worker(threading.Thread):

    def run(self):
        w_logger = logging.getLogger(self.name)
        i = 0
        ns.val = 0

        for i in range(stop):
            ns.val += 1
            i+=1
            w_logger.debug(f"Thread: {self.name}, value: {ns.val}")
            value = get_threadlocal_var(ns, "val")
            assert i == value
        value = get_threadlocal_var(ns, "val")
        assert stop==value


def test_get_threadlocal_var_exist_different_thread(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    w1 = Worker()
    w2 = Worker()
    w1.start()
    w2.start()
    w1.join()
    w2.join()

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'val')

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'nonexist')


def test_del_threadlocal_var_empy(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    del_threadlocal_var(thread_locals, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


def test_del_threadlocal_var_exist(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    thread_locals.value = 1

    del_threadlocal_var(thread_locals, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(ns, 'value')

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')

class Box(object):

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


def test_threadlocal_var_empty(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    expValue = 5
    box = threadlocal_var(thread_locals, 'value', Box, value=expValue)
    value = box.value
    assert expValue == value

    box2 = get_threadlocal_var(thread_locals, 'value')
    value2 = box2.value
    assert expValue == value2

    assert box == box2

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')


def test_threadlocal_var_exists(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    thread_locals = threading.local()

    expValue = 5
    box = threadlocal_var(thread_locals, 'value', Box, value=expValue)
    value = box.value
    assert expValue == value

    box2= threadlocal_var(thread_locals, 'value', Box, value=100)
    value2 = box2.value
    assert expValue == value2

    assert box == box2

    box3 = threadlocal_var(thread_locals, 'value', Box)
    value3 = box3.value
    assert expValue == value3

    assert box == box3

    with pytest.raises(ValueError):
        get_threadlocal_var(thread_locals, 'nonexist')







if __name__ == "__main__":
    pytest.main([__file__])



import logging
import pytest
import inspect
from alexber.utils.inspects import issetdescriptor, ismethod, has_method
from collections import OrderedDict

logger = logging.getLogger(__name__)


class Example(object):
    def __init__(self, name, address=None, *args, **kwargs):
        pass

    def foo1(self):
        pass

    @staticmethod
    def method1():
        pass

    @classmethod
    def method2(cls):
        pass

    @property
    def att1(self):
        return 5

    @att1.setter
    def att1(self, value):
        pass

    @property
    def att_get_only(self):
        return 5

    def _set_only(self):
        pass

    att_set_only = property(fset=_set_only)


    def set_x(self, x):
        pass

    def baz(self, a, b, c):
        pass

    def seed(self, a=None, version=2):
        pass

class Base(object):

    @property
    def att1(self):
        return 5

    @att1.setter
    def att1(self, value):
        pass

    @property
    def att_get_only(self):
        return 5

    def set_x(self, x):
        pass

class Derived(Base):
    def __init__(self, name, address=None, *args, **kwargs):
        pass



@pytest.mark.parametrize(
     'obj',
    (Example,
     Derived),
)
def test_property_get_set(request, obj):
    logger.info(f'{request._pyfuncitem.name}()')
    results = inspect.getmembers(obj, predicate=issetdescriptor)
    d = {key: value for key, value in results}

    prop = d['att1']

    setter = prop.fset
    setter(self=None, value=100)

    prop = d['att_get_only']
    setter = prop.fset
    assert setter is None


@pytest.mark.parametrize(
     'f',
    (Example,
     Derived),
)
def test_signature(request, f):
    logger.info(f'{request._pyfuncitem.name}()')
    sig = inspect.signature(f)

    d = OrderedDict()
    for var in sig.parameters.keys():
        d[var] = None

    var = d['name']

    logger.debug(d)
    logger.debug(var)

@pytest.mark.parametrize(
     'f',
    (Example,
     Derived),
)
#see https://stackoverflow.com/questions/218616/getting-method-parameter-names-in-python/45781963#45781963
#see https://stackoverflow.com/questions/218616/getting-method-parameter-names-in-python/44261531#44261531
def test_binding(request, f):
    logger.info(f'{request._pyfuncitem.name}()')
    d = OrderedDict()
    d['name'] = 'metoo'

    sig = inspect.signature(f)
    bound_args = sig.bind(**d)
    bound_args.apply_defaults()
    kwargs = bound_args.arguments
    obj = f(**kwargs)
    assert obj is not None

class SubDerived(Derived):
    def set_x(self, x):
        pass

class Derivitive(SubDerived, Base):
    def set_x(self, x):
        pass

    def cool(self):
        pass

@pytest.mark.parametrize(
     'cls, method_name, exp_value',

    [
        (Example, '__init__', True),
        (Example, 'foo1', True),
        (Example, 'method1', True),
        (Example, 'method2', True),
        (Example, 'att1', True),
        (Example, 'att_get_only', True),
        (Example, '_set_only', True),
        (Example, 'set_x', True),
        (Example, 'baz', True),
        (Example, '__init__', True),
        (Example, 'seed', True),
        (Example, 'this_is_method_is_not_here', False),
        (Derived, '__init__', True),
        (SubDerived, 'set_x', True),
        (SubDerived, 'set_non_exists', False),
        (Derivitive, 'set_x', True),
        (Derivitive, 'cool', True),
        (Derivitive, 'set_non_exists', False),


    ]
)
def test_has_method(request, cls, method_name, exp_value):
    logger.info(f'{request._pyfuncitem.name}()')

    b = has_method(cls, method_name)
    assert exp_value == b

if __name__ == "__main__":
    pytest.main([__file__])



#inspired by https://stackoverflow.com/questions/1408171/thread-local-storage-in-python


def threadlocal_var(thread_locals, varname, factory, *args, **kwargs):
  v = getattr(thread_locals, varname, None)
  if v is None:
    v = factory(*args, **kwargs)
    setattr(thread_locals, varname, v)
  return v

def get_threadlocal_var(thread_locals, varname):
    v = threadlocal_var(thread_locals, varname, lambda : None)
    if v is None:
        raise ValueError(f"threadlocal's {varname} is not initilized")
    return v

def del_threadlocal_var(thread_locals, varname):
    try:
        delattr(thread_locals, varname)
    except AttributeError:
        pass





from enum import *
import enum #don't delete
import logging
logger = logging.getLogger(__name__)


_orig_enum_new = Enum.__new__

#we want to use None to indicate missing Enumm, but
#https://bugs.python.org/issue34536 prohibts using None in _missing_() @classmethod
#this fix re-enable it
def _fixed_new_enum(cls, value):
    try:
        return _orig_enum_new(cls, value)
    except ValueError as e:
        msg = str(e)
        if 'is not a valid' in msg:
            return None
        raise e

Enum.__new__ = _fixed_new_enum

class StrAsReprMixinEnum(Enum):
    '''
    This is Enum Mixin that has __str__() equal to __repr__().
    '''
    def __str__(self):
        return self.__repr__()


class AutoNameMixinEnum(Enum):
    '''
    This is Enum Mixin that generate value equal to the name.
    '''
    def _generate_next_value_(name, start, count, last_values):
        return name



class MissingNoneMixinEnum(Enum):
    '''
    This is Enum Mixin will return None if value will not be found.
    '''

    @classmethod
    def _missing_(cls, value):
        # raise ValueError("%r is not a valid %s" % (value, cls.__name__))
        return None


class LookUpMixinEnum(StrAsReprMixinEnum, MissingNoneMixinEnum):
    '''
    This is Enim Mixin that is designed to be used for lookup by value.
    If lookup fail, None will be return.
    Also, __str__() will return the same value as __repr__()
    '''
    pass





import logging
import pytest

from alexber.utils import StrAsReprMixinEnum, LookUpMixinEnum, AutoNameMixinEnum, MissingNoneMixinEnum, Enum
from alexber.utils.enums import enum, auto

# #use standard Enum.__new__
def _orig_enum_new(cls, value):
    # all enum instances are actually created during class construction
    # without calling this method; this method is called by the metaclass'
    # __call__ (i.e. Color(3) ), and by pickle
    if type(value) is cls:
        # For lookups like Color(Color.RED)
        return value
    # by-value search for a matching enum member
    # see if it's in the reverse mapping (for hashable values)
    try:
        if value in cls._value2member_map_:
            return cls._value2member_map_[value]
    except TypeError:
        # not there, now do long search -- O(n) behavior
        for member in cls._member_map_.values():
            if member._value_ == value:
                return member
    # still not found -- try _missing_ hook
    try:
        exc = None
        result = cls._missing_(value)
    except Exception as e:
        exc = e
        result = None

    if isinstance(result, cls):
        return result
    else:
        ve_exc = ValueError("%r is not a valid %s" % (value, cls.__name__))
        if result is None and exc is None:
            raise ve_exc
        elif exc is None:
            exc = TypeError(
                'error in %s._missing_: returned %r instead of None or a valid member'
                % (cls.__name__, result)
            )
        exc.__context__ = ve_exc
        raise exc


Enum.__new__ = _orig_enum_new
StrAsReprMixinEnum.__new__ = _orig_enum_new
LookUpMixinEnum.__new__ = _orig_enum_new
AutoNameMixinEnum.__new__ = _orig_enum_new
MissingNoneMixinEnum.__new__ = _orig_enum_new


logger = logging.getLogger(__name__)

#
# @pytest.fixture()
# def my_dependency():
#     return 42
#
# #@pytest.mark.ws
# def test_first(my_dependency ):
#     logger.info('Testing first')
#


class TestStandardEnum(object):
    @enum.unique
    class Color1(Enum):
        RED = 10
        BLUE = 20
        GREEN = 30

    @enum.unique
    class Color2(Enum):
        RED = 'r'
        BLUE = 'b'
        GREEN = 'g'

    @enum.unique
    class Color3(Enum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()


    @pytest.mark.parametrize(
        'cls,is_by_name,value',
        [
            (Color1, True, 10),
            # (Color1, False, 'RED'),
            # (Color2, True, 'b'),
            # (Color2, False, 'BLUE'),
            # (Color3, True, 3),
            # (Color3, False, 'GREEN'),
        ]
    )
    def test_builtin_enumuration(self, request, cls, is_by_name, value):
        logger.info(f'{request._pyfuncitem.name}()')
        some_enum = cls(value) if is_by_name else cls[value]
        logger.info(f'name is {some_enum!r}')
        sec_enum = cls[some_enum.name]
        assert some_enum == sec_enum

        logger.debug(some_enum)
        logger.debug(f'repr is {some_enum!r}')
        logger.debug(f'str is {some_enum!s}')
        [logger.debug(f'{member!r}') for member in cls.__members__.values()]

        with pytest.raises(KeyError) as excinfo:
            cls['missing_key_here_it_is']
            # logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))

        with pytest.raises(ValueError) as excinfo:
            cls('missing_value_here_it_is')
            # logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))


class TestStrAsReprMixinEnum(object):

    @enum.unique
    class ColorStr(StrAsReprMixinEnum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()

    @enum.unique
    class ColorStr2(StrAsReprMixinEnum, Enum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()



    @pytest.mark.parametrize(
         'cls',
        [
            ColorStr,
            ColorStr2
        ])
    def test_StrAsReprMixinEnum(self, request, cls):
        logger.info(f'{request._pyfuncitem.name}()')
        some_enum = cls(1)
        logger.info(f'name is {some_enum!r}')
        sec_enum = cls[some_enum.name]
        assert some_enum==sec_enum

        logger.debug(some_enum)
        logger.debug(f'repr is {some_enum!r}')
        logger.debug(f'str is {some_enum!s}')
        [logger.debug(f'{member!r}') for member in cls.__members__.values()]

        assert repr(some_enum) == str(some_enum)


class TestAutoNameMixinEnum(object):
    @enum.unique
    class ColorAutoName(AutoNameMixinEnum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()

    @enum.unique #AutoNameMixinEnum has to be last (see EnumMeta.__prepare__())
                 #method _generate_next_value_() will be taken from theire
    class ColorAutoName2(StrAsReprMixinEnum, AutoNameMixinEnum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()

    @enum.unique #AutoNameMixinEnum has to be last (see EnumMeta.__prepare__())
    class ColorAutoNameWrong(AutoNameMixinEnum, StrAsReprMixinEnum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()

    @enum.unique #AutoNameMixinEnum has to be last (see EnumMeta.__prepare__())
    class ColorAutoNameWrong2(AutoNameMixinEnum, Enum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()




    @pytest.mark.parametrize(
         'cls',
        [
            ColorAutoName,
            ColorAutoName2,
        ])
    def test_AutoNameMixinEnum(self, request, cls):
        logger.info(f'{request._pyfuncitem.name}()')
        some_enum = cls('RED')
        logger.info(f'name is {some_enum!r}')
        sec_enum = cls[some_enum.name]
        assert some_enum==sec_enum

    @pytest.mark.parametrize(
         'cls,is_rep_check',
        [
            (ColorAutoNameWrong, True),
            (ColorAutoNameWrong2, False),
        ])

    def test_AutoNameWrongMixinEnum(self, request, cls, is_rep_check):
        logger.info(f'{request._pyfuncitem.name}()')

        with pytest.raises(ValueError) as excinfo:
            some_enum = cls('RED')
            logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))

        some_enum = cls(1)
        logger.info(f'name is {some_enum!r}')
        sec_enum = cls[some_enum.name]
        assert some_enum==sec_enum

        logger.debug(some_enum)
        logger.debug(f'repr is {some_enum!r}')
        logger.debug(f'str is {some_enum!s}')
        [logger.debug(f'{member!r}') for member in cls.__members__.values()]

        if is_rep_check:
            assert repr(some_enum) == str(some_enum)


class TestMissingNoneMixinEnum(object):

    @enum.unique
    class ColorMissingNone(MissingNoneMixinEnum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()


    def test_MissingNoneEnum(self, request):
        logger.info(f'{request._pyfuncitem.name}()')

        some_enum = TestMissingNoneMixinEnum.ColorMissingNone(1)
        logger.info(f'name is {some_enum!r}')
        sec_enum = TestMissingNoneMixinEnum.ColorMissingNone[some_enum.name]
        assert some_enum==sec_enum

        with pytest.raises(KeyError) as excinfo:
            TestMissingNoneMixinEnum.ColorMissingNone[10_000]
            #logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))

        with pytest.raises(ValueError) as excinfo:
            TestMissingNoneMixinEnum.ColorMissingNone(10_000)



class TestLookUpMixinEnum(object):

    @enum.unique
    class ColorLookUp(LookUpMixinEnum):
        RED = 'r'
        BLUE = 'b'
        GREEN = 'g'

    @enum.unique
    class ColorLookUpAuto(LookUpMixinEnum, AutoNameMixinEnum):
        RED = auto()
        BLUE = auto()
        GREEN = auto()



    def test_ColorLookUp(self, request):
        logger.info(f'{request._pyfuncitem.name}()')

        some_enum = TestLookUpMixinEnum.ColorLookUp('r')
        logger.info(f'name is {some_enum!r}')
        sec_enum = TestLookUpMixinEnum.ColorLookUp[some_enum.name]
        assert some_enum == sec_enum

        with pytest.raises(KeyError) as excinfo:
            TestLookUpMixinEnum.ColorLookUp['missing_value_here_it_is']
            # logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))
        with pytest.raises(ValueError) as excinfo:
            TestLookUpMixinEnum.ColorLookUp('missing_value_here_it_is')


    def test_ColorLookUpAuto(self, request):
        logger.info(f'{request._pyfuncitem.name}()')

        some_enum = TestLookUpMixinEnum.ColorLookUpAuto('RED')
        logger.info(f'name is {some_enum!r}')
        sec_enum = TestLookUpMixinEnum.ColorLookUpAuto[some_enum.name]
        assert some_enum == sec_enum

        with pytest.raises(KeyError) as excinfo:
            TestLookUpMixinEnum.ColorLookUpAuto['missing_value_here_it_is']
            # logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))
        with pytest.raises(ValueError) as excinfo:
            TestLookUpMixinEnum.ColorLookUpAuto('missing_value_here_it_is')



class TestLookUpAutoNameMixinEnum(object):

    def addComp(enumeration):
        """Class decorator for enumerations """
        all = []
        for member in enumeration.__members__.values():
            all.append(member.value)

        enumeration.comp = all
        return enumeration

    @enum.unique
    @addComp
    class Color(LookUpMixinEnum, AutoNameMixinEnum):
        RED = 'R'
        BLUE = 'B'
        GREEN = "G"

    def test_Color(self, request):
        logger.info(f'{request._pyfuncitem.name}()')

        some_enum = TestLookUpAutoNameMixinEnum.Color('R')
        logger.info(f'name is {some_enum!r}')
        sec_enum = TestLookUpAutoNameMixinEnum.Color[some_enum.name]
        assert some_enum == sec_enum

        with pytest.raises(KeyError) as excinfo:
            TestLookUpAutoNameMixinEnum.Color['missing_value_here_it_is']
            # logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))
        with pytest.raises(ValueError) as excinfo:
             TestLookUpAutoNameMixinEnum.Color('missing_value_here_it_is')

        value = TestLookUpAutoNameMixinEnum.Color.comp
        assert value is not None



if __name__ == "__main__":
    pytest.main([__file__])




"""
This module adopts its behavior dependent on availability of Python packages.

This module optionally depends on ymlparseser module.
Method format_template() is used in emails module.

Note: This module will work if you have only standard Python package. You just can't change delimiters values.
Note: API and implementation of this module is unstable and can change without prior notice.
"""

import warnings

def format_template(template, **kwargs):
    """
    This is main method of this module.
    Note: API of this method is unstable and can change without prior notice.

    Template is expected to be compatible with Jinja2 one.

    Current implementation make delimiters compatible with str.format() and use it.


    :param template: str, typically with {{my_variable}}
    :param jinja2ctx:  Jinja2 Environment that is consulted what is delimiter for variable's names.
                       if is not provided, HiYaPyCo.jinja2ctx is used. See ymlparsers.initConfig().
                       if is not provided, than defaults are used (see jinja2.defaults).
    :param jinja2Lock: Lock to be used to atomically get variable_start_string and variable_end_string from jinja2ctx.
                       if is not provided, HiYaPyCo.jinja2Lock is used.. See ymlparsers.initConfig().
    :return: fromated str
    """
    if template is None:
        return None
    s = _convert_template_to_string_format(template, **kwargs)
    ret = s.format(template, **kwargs)
    return ret


try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=r'.*?yaml*?', module=r'.*?ymlparsers.*?')
    from . ymlparsers import HiYaPyCo
    _isHiYaPyCoAvailable = True
except ImportError:
    _isHiYaPyCoAvailable = False

_a1 = None
_a2 = None
try:
    try:
        from jinja2.defaults import VARIABLE_START_STRING as _a1, VARIABLE_END_STRING as _a2
        _isJinja2DefaultAvailable = True
    except ImportError:
        try:
            from jinja2.environment import VARIABLE_START_STRING as _a1, VARIABLE_END_STRING as _a2
            _isJinja2DefaultAvailable = True
        except ImportError:
            _isJinja2DefaultAvailable = False
finally:
    del _a1
    del _a2

_VARIABLE_START_STRING = None
_VARIABLE_END_STRING = None


def _init_globals():
    """
    This method is called during module import.
    This method is idempotent.
    """
    global _VARIABLE_START_STRING, _VARIABLE_END_STRING

    if _isJinja2DefaultAvailable:
        p1 = None
        p2 = None
        try:
            from jinja2.defaults import VARIABLE_START_STRING as p1, VARIABLE_END_STRING as p2
        except ImportError:
            from jinja2.environment import VARIABLE_START_STRING as p1, VARIABLE_END_STRING as p2

        if p1 is None or p2 is None:
            raise ImportError('VARIABLE_START_STRING or VARIABLE_END_STRING are not defined')

        _VARIABLE_START_STRING = p1
        _VARIABLE_END_STRING = p2
    else:
        _VARIABLE_START_STRING = '{{'
        _VARIABLE_END_STRING = '}}'


_init_globals()

def _normalize_var_name(text, start_del, end_del):
    """
    Search&replace all pairs of (start_del, end_del) with pairs of ({, }).

    :param text: str to normalize
    :param start_del: delimiter that indicates start of variable name, typically {{
    :param end_del: delimiter that indicates end of variable name, typically }}
    :return:
    """

    if start_del is None or start_del not in text or end_del not in text:
        return text

    first_ind = 0
    len_end_del = len(end_del)

    while True:
        first_ind = text.find(start_del, first_ind)
        if first_ind < 0:
            break
        last_ind =  text.find(end_del, first_ind)
        var = text[first_ind:last_ind+len_end_del]
        var = var.replace('.', '_')
        #text[first_ind:last_ind] = var
        text = text[:first_ind]+var+text[last_ind+len_end_del:]
        first_ind = last_ind+len_end_del
    return text


def __convert_template_to_string_format(template, **kwargs):
    """
    This is utility method that make template usable with string format.


    :param template: str, typically with {{my_variable}}
    :param default_start: Typically {{ but can be any other delimiter that points to start of the token variable.
    :param default_end:   Typically }} but can be any other delimiter that points to end of the token variable.
    :return: template: str with {my_variable}
    """
    if template is None:
        return None

    default_start = kwargs.pop('default_start', None)
    default_end = kwargs.pop('default_end', None)

    template = _normalize_var_name(template, default_start, default_end)

    ret = template.replace(f'{default_start} ', f'{default_start}') \
        .replace(f'{default_start}', '{') \
        .replace(f' {default_end}', f'{default_end}') \
        .replace(f'{default_end}', '}')
    return ret

def _convert_template_to_string_format(template, **kwargs):
    """
    This is utility method that make template usable with string format.

    if both jinja2ctx and jinja2Lock are provided, than they are used to determine various delimiters
    (jinja2Lock is used to read the values from jinja2ctx atomically).

    if both jinja2ctx and jinja2Lock are not provided, than
        If ymlparsers is usable (it's 3rd party dependencies are available, one if each is jinja2)
        than it's jinja2ctx (Jinja2's Environment) will be consulted for the various delimiters.
        Otherwise, if jinja2 is available than we will use it's defaults for constricting Jinja2's Environment
        for the various delimiters.
        Otherwise, some sensible defaults (default values from some version of Jinja2) will be used.

    You can't provide jinja2Lock without providing jinja2ctx (you can't provide your jinja2Lock for HiYaPyCo.jinja2ctx).

    You can provide jinja2ctx without jinja2Lock. Than you will give up atomicity for determining various delimiters.

    :param template: str, typically with {{my_variable}}
    :param jinja2ctx:  Jinja2 Environment that is consulted what is delimiter for variable's names.
                       if is not provided, HiYaPyCo.jinja2ctx is used. See ymlparsers.initConfig().
                       if is not provided, than defaults are used (see jinja2.defaults).
    :param jinja2Lock: Lock to be used to atomically get variable_start_string and variable_end_string from jinja2ctx.
                       if is not provided, HiYaPyCo.jinja2Lock is used.. See ymlparsers.initConfig().
    :return: template: str with {my_variable}
    """

    if template is None:
        return None

    jinja2ctx = kwargs.pop('jinja2ctx', None)
    jinja2Lock = kwargs.pop('jinja2Lock', None)

    if _isHiYaPyCoAvailable and jinja2ctx is None and jinja2Lock is not None:
        raise ValueError("You can't provide your jinja2Lock for HiYaPyCo.jinja2ctx")

    if _isHiYaPyCoAvailable and jinja2ctx is None:
        jinja2ctx  = HiYaPyCo.jinja2ctx
        jinja2Lock = HiYaPyCo.jinja2Lock    #we should use HiYaPyCo.jinja2Lock for HiYaPyCo.jinja2ctx

    #default_start, default_end
    if jinja2ctx is None:
        if jinja2Lock is None:
            default_start = _VARIABLE_START_STRING
            default_end = _VARIABLE_END_STRING
        else:
            with jinja2Lock:
                default_start = _VARIABLE_START_STRING
                default_end = _VARIABLE_END_STRING

    else:
        if _isHiYaPyCoAvailable and HiYaPyCo.jinja2ctx is not None and HiYaPyCo.jinja2Lock is None:
            raise ValueError('HiYaPyCo.jinja2ctx is not None, but HiYaPyCo.jinja2Lock is None')

        if jinja2Lock is None:
            # jinja2ctx was provided, but jinja2Lock wasn't, it is ok
            # (maybe jinja2ctx is local variable?)
            default_start = jinja2ctx.variable_start_string
            default_end = jinja2ctx.variable_end_string
        else:
            with jinja2Lock:
                default_start = jinja2ctx.variable_start_string
                default_end = jinja2ctx.variable_end_string

    ret = __convert_template_to_string_format(template, default_start=default_start, default_end=default_end)
    return ret





import pytest
import logging

logger = logging.getLogger(__name__)
from alexber.utils.importer import importer, new_instance

class TestImporter(object):
    def test_imported(self, request):
        logger.info(f'{request._pyfuncitem.name}()')

        cls_name = 'pathlib.Path'
        kls = importer(cls_name)

        path = kls()
        logger.info(path.resolve())

    def test_imported_implicit_namespace_package(self, request):
        logger.info(f'{request._pyfuncitem.name}()')

        module_name = 'tests.utils.splitpackage.other_module'
        module = importer(module_name)
        value = module.theansweris()
        evp_value = 42
        pytest.assume(evp_value, value)



def test_new_instance_function(request):
    logger.info(f'{request._pyfuncitem.name}()')
    cls_name = 'pathlib.Path.cwd'
    kls = new_instance(cls_name)

    path = kls()
    logger.info(path.absolute())


class PlayerEmpty(object):
    pass


class PlayerInitFull(object):
    def __init__(self, *args, **kwargs):
        pass

class PlayerInitArg(object):
    def __init__(self, first_name):
        self.first_name = first_name

class PlayerInitDefaultArg(object):
    def __init__(self, first_name='John', **kwargs):
        self.first_name = first_name

class PlayerNewAndInitEmpty(object):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        logger.debug("PlayerNewAndInitEmpty.__new__()")
        return self

class PlayerNewOnlyEmpty(object):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        return self

    def __init__(self, *args, **kwargs):
        pass

class PlayerNewOnlyEmptySneaky(object):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        return 3    #3 is not instace of this class '__init__" shouldn't be called

    def __init__(self, *args, **kwargs):
        pass




from abc import ABCMeta, abstractmethod

class PlayerAbstractWithAbstractMethod(object, metaclass=ABCMeta):
    @abstractmethod
    def foo(self):
        raise NotImplementedError


class PlayerAbstractEmptyWithoutAbstractMethod(object, metaclass=ABCMeta):
    pass

class PlayerAbstractFullWithoutAbstractMethod(object, metaclass=ABCMeta):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        return self

    def __init__(self, *args, **kwargs):
        pass



def test_new_instance_arg(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mock = mocker.spy(PlayerInitArg, '__init__')

    input = '.'.join([__name__, PlayerInitArg.__name__])
    first_name = 'John'
    player=new_instance(input, first_name)
    assert player.first_name == first_name
    assert player.__init__.call_count == 1


def test_new_instance_default_arg1(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    input = '.'.join([__name__, PlayerInitDefaultArg.__name__])
    first_name = 'John'
    player=new_instance(input, first_name)
    assert player.first_name == first_name

def test_new_instance_default_arg2(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')

    input = '.'.join([__name__, PlayerInitDefaultArg.__name__])
    player=new_instance(input)
    assert player.first_name == "John"

def test_new_instance_default_arg3(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mock = mocker.spy(PlayerInitDefaultArg, '__init__')

    input = '.'.join([__name__, PlayerInitDefaultArg.__name__])
    player=new_instance(input, last_name='Google')
    assert player.first_name == "John"


@pytest.mark.parametrize(
     'plcls',
    (PlayerEmpty, PlayerInitFull,PlayerNewOnlyEmpty, PlayerNewOnlyEmpty,PlayerNewAndInitEmpty,
     PlayerAbstractEmptyWithoutAbstractMethod, PlayerAbstractFullWithoutAbstractMethod),
)
def test_new_instance(request, plcls):
    logger.info(f'{request._pyfuncitem.name}()')

    input = '.'.join([__name__, plcls.__name__])
    player = new_instance(input)
    assert player is not None


def test_new_instance_init_subclass(request):
    logger.info(f'{request._pyfuncitem.name}()')
    module_name, _ = __name__.rsplit(".", 1)
    input = '.'.join([module_name, 'method_overloading_test', 'PlayerAustralianPhilosopher'])

    player=new_instance(input)
    assert player is not None


def test_new_instance_abstract_method(request):
    logger.info(f'{request._pyfuncitem.name}()')
    input = '.'.join([__name__, PlayerAbstractWithAbstractMethod.__name__])

    with pytest.raises(TypeError, match='abstract method') as excinfo:
        player = new_instance(input)
    logger.debug(excinfo.value, exc_info=(excinfo.type, excinfo.value, excinfo.tb))



def test_new_instance_sneaky(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    mock = mocker.spy(PlayerNewOnlyEmptySneaky, '__init__')

    input = '.'.join([__name__, PlayerNewOnlyEmptySneaky.__name__])
    player=new_instance(input)
    assert not hasattr(player.__init__, 'call_count') or player.__init__.call_count == 0


if __name__ == "__main__":
    pytest.main([__file__])


import logging
import pytest

logger = logging.getLogger(__name__)

from uuid import uuid1
_int_from_bytes = int.from_bytes
from os import urandom
from alexber.utils import uuid1mc

#refference implementation
#https://stackoverflow.com/questions/1785503/when-should-i-use-uuid-uuid1-vs-uuid-uuid4-in-python
def ref_uuid1mc():
    return uuid1(_int_from_bytes(urandom(6), "big") | 0x010000000000)

def test_uuid1mc(request, mocker):
    logger.info(f'{request._pyfuncitem.name}()')
    # logger.debug(urandom(6))
    rnd = b'W\x91\x8e\xb3!\x08'

    mocker.patch("random._urandom",  side_effect=lambda _: rnd, create=True)
    mocker.patch('.'.join([__name__, 'urandom']),  side_effect=lambda _: rnd, create=True)
    mocker.patch('.'.join([__name__, 'uuid1']), side_effect=lambda p:p, autospec=True, spec_set=True)
    mocker.patch('.'.join(['alexber.utils.uuids', '_uuid1']), side_effect=lambda p: p, autospec=True, spec_set=True)

    exp = ref_uuid1mc()
    actual = uuid1mc()
    assert  exp==actual





if __name__ == "__main__":
    pytest.main([__file__])



"""
This module contains extensions of the logging handlers.
See https://docs.python.org/3/library/logging.handlers.html#module-logging.handlers for more details.

See here https://medium.com/analytics-vidhya/my-emails-module-3ad36a4861c5 for documentation.
All handlers are thread safe.

It is better to use EmailStatus context manager with configured emailLogger.
See docstring of EmailStatus.

This module optionally depends on ymlparseser module.

If you want to change delimeters used to indicate variable declaration inside template, see docstring of the
OneMemoryHandler.get_subject() method.

"""

import io as _io
import warnings
import logging
from logging.handlers import SMTPHandler as _logging_SMTPHandler
from logging.handlers import MemoryHandler as _logging_MemoryHandler
from smtplib import SMTP as _SMTP
import contextlib
from collections.abc import Mapping
from email.message import EmailMessage as _EmailMessage
from email.policy import SMTPUTF8 as _SMTPUTF8
from ..utils import threadlocal_var
from . importer import importer
from .parsers import is_empty
from . _ymlparsers_extra import format_template as _format_template

FINISHED = logging.FATAL+10
logging.addLevelName(FINISHED, 'FINISHED')

default_smtp_cls = None
default_smtp_port = None

import threading
_thread_locals = threading.local()



class SMTPHandler(_logging_SMTPHandler):
    """
    It's purpose is to connect to SMTP server and actually send the e-mail.
    This class expects for record.msg to be built EmailMessage.
    You can also change use of underline smtplib.SMTP class to smtplib.SMTP_SSL, smtplib.LMTP, or any another class.

    This implementation is Thread safe.
    """
    def __init__(self, *args, **kwargs):
        smtp_cls_name = kwargs.pop('smtpclsname', None)
        if isinstance(smtp_cls_name, str):
            smtp_cls = importer(smtp_cls_name)
        else:
            smtp_cls = smtp_cls_name
        if smtp_cls is None:
            smtp_cls = default_smtp_cls
        self.smtp_cls = smtp_cls

        super().__init__(*args, **kwargs)


    def emit(self, record):
        """
        Emit a record.

        Format the record and send it to the specified addressees.
        """
        try:
            msg = record.msg

            if msg is not None:
                port = self.mailport
                if not port:
                    port = getattr(self.smtp_cls, 'default_port', None)

                if port is None:
                    port = default_smtp_port

                if port is None:
                    raise ValueError("{self.smtp_cls} class that will be used doesn't contain default_port field."
                                     "You should explicitly specify default_smtp_port.")


                smtp = self.smtp_cls(self.mailhost, port, timeout=self.timeout)

                if self.username:
                    if self.secure is not None:
                        smtp.ehlo()
                        smtp.starttls(*self.secure)
                        smtp.ehlo()
                    smtp.login(self.username, self.password)
                smtp.send_message(msg, self.fromaddr, self.toaddrs)
                smtp.quit()
        except Exception:
            self.handleError(record)

class BaseOneMemoryHandler(_logging_MemoryHandler):
    """
    This implementation is Thread safe. Each Thread aggregates it's own buffer of the log messages.
    """

    def __init__(self, *args, **kwargs):
        subject = kwargs.pop('subject')
        self.subject = subject

        self.createLock()
        lock = self.lock

        self.acquire()
        try:
            super().__init__(*args, **kwargs)
            #restore lock
            self.lock=lock
            del self.buffer
            threadlocal_var(_thread_locals, 'buffer', lambda: [])
        finally:
            self.release()

    def shouldFlush(self, record):
        """
        Check for buffer full or a record at the flushLevel or higher.
        """
        return record.levelno >= self.flushLevel

    def calc_msg_params(self, *args, **kwargs):
        """
        Hook to be overridden in the sub-class.
        """
        pass

    def get_subject(self, *args, **kwargs):
        """
        You may override this method.
        """
        ret = self.subject
        return ret

    def create_one_record(self, records):
        length = len(records)
        last_record = records[-1] if length > 0 else None
        is_finished = False if last_record is None else self.shouldFlush(last_record)

        self.calc_msg_params(records=records,
                             is_finished=is_finished,
                             last_record=last_record
                             )

        with _io.StringIO() as body:
            for record in records:
                message = self.format(record)
                message = message.replace('\n', '<br />'+'\n')
                body.write(message)
                body.write("<br />")
                body.write('\n')


            msg = _EmailMessage(policy=_SMTPUTF8)

            msg['Subject'] = self.get_subject()
            msg.set_content(body.getvalue(), subtype='html', cte='8bit')

        last_record.msg = msg
        return last_record

    def emit(self, record):
        """
        Emit a record.

        Append the record. If shouldFlush() tells us to, call flush() to process
        the buffer.
        """
        buffer = threadlocal_var(_thread_locals, 'buffer', lambda: [])


        buffer.append(record)
        if self.shouldFlush(record):
            self.flush()


    def flush(self):
        """
        For a MemoryHandler, flushing means just sending the buffered
        records to the target, if there is one. Override if you want
        different behaviour.

        The record buffer is also cleared by this operation.
        """
        self.acquire()


        try:
            buffer = threadlocal_var(_thread_locals, 'buffer', lambda: [])

            if self.target and buffer:
                # for record in self.buffer:
                #     self.target.handle(record)

                record = self.create_one_record(buffer)
                self.target.handle(record)

                #self.buffer = []

                setattr(_thread_locals, 'buffer', [])
        finally:
            self.release()



class OneMemoryHandler(BaseOneMemoryHandler):
    """
    This is variant of logging.handlers.MemoryHandler.

    This implementation is Thread safe.

    This handler aggregates log messages until FINISHED log-level is received or we've figure out that application is
    going to terminate abruptly and we have some log messages in the buffer (see below).

    On such event all messages (in the current Thread) are aggregated to the single EmailMessage.
    The subject of the EmailMessage is determined by get_subject() method (see it's docstring for the details).

    There are cases on which FINISHED log-level is not received, but we know that application is going to terminate
    abruptly and we did aggregated some messages. See docstring of the calc_abrupt_vars() method for the details.
    On such a cases we will send EmailMessage as if FINISHED log-level was recieved, but using abruptvars field
    that was set in the constructor in calc_abrupt_vars() method.

    If you want to change delimeters used to indicate variable declaration inside template, see docstring of the
    get_subject() method.

    Note: It is better to use EmailStatus context manager with configured emailLogger. See docstring of EmailStatus.

    """
    DEFAULT_ABRUPT_VARS =  {'status': 'Finished Abruptly'}

    def calc_msg_params(self, *args, **kwargs):
        records = kwargs['records']
        is_finished = kwargs['is_finished']
        last_record = kwargs['last_record']

        if is_finished:
            variables = last_record.msg
            b = is_empty(variables)
            if b:
                variables = {}
            else:
                if not isinstance(variables, Mapping):
                    raise ValueError("Only Mapping (dict) is expected as last log message")

            self.variables = variables
        else:
            self.variables = self.abruptvars


        if is_finished:
            records.pop()

    format_template = staticmethod(_format_template)

    def get_subject(self, *args, **kwargs):
        """
        You may override this method.

        If your subject doesn't contains place holders, for example, 'Aggregates log from the Demo application',
        there is no need to override this method. See also docstring of EmailStatus context manager.

        If you want to change delimeters used to indicate variable declaration inside template, you have following
        options:

        1. You can override this method and pass Jinja2 Environment object to format_template() method.
            If your application is mutli-threaded, this object should be non-shareble (maybe locally defined).
            If this object is shared, see another options below.
        2. You can override this method and pass Jinja2 Environment object with threading.RLock (or any other lock)
        object in order to ensure atomic read of all decimetres.
        3. Call alexber.utils.ymplparser.initConfig() method to initialize HiYaPyCo.jinja2ctx (together will
        HiYaPyCo.jinja2Lock). It will be consulted for delimeters in format_template() method. Please, note:
            This object will be used by also by another modules (init_app_conf.py for example), so this change
            may effect also another parts of your applications if you use when.
            You're changing global variable, please make sure, you own code is aware of this change.
            You should ensure that dependencies that are required by ymplparser module are installed.
            See docstring of ymplparser module for mode details.
        4. You can override this method with your custom logic.

        """
        subject = self.subject
        ret = self.format_template(subject, **self.variables)
        return ret

    def calc_abrupt_vars(self, *args, **kwargs):
        """
        You can override this method.

        If your subject doesn't contains place holders, for example, 'Aggregates log from the Demo application',
        there is no need to override this method.

        There are alternatives though to change default behaviour of this method
        (see below).

        This method need to define self.abruptvars field.

        If your application is stopped abruptly (for example, by SIGINT) and your buffer (in the current Thread)
        has some messages,
        logging system at shutdown will flush and close all handlers, so the buffer without message
        with log-level FINISHED (and hence without variable substitution kwargs) will be flushed.
        self.abruptvars will be used instead.

        If you application has execution path that doesn't ends by sending log messages with log-level FINISHED
        (for example, you have uncaught exception that cause your application to terminate abruptly)
        and your buffer (in the current Thread) has some messages,
        logging system at shutdown will flush and close all handlers, so the buffer without without message
        with log-level FINISHED (and hence without variable substitution kwargs) will be flushed.
        self.abruptvars will be used instead.
        It is recommended to use EmailStatus context manager in order to avoid this.

        The default implementation checks if abruptvars kwargs was supplied in constructor. If it did, it will be used
        put as self.abruptvars. If it doesn't DEFAULT_ABRUPT_VARS is used as self.abruptvars.

        :param abruptvars: kwargs to be put as self.abruptvars. Optional.

        """
        abruptvars = kwargs.pop('abruptvars', None)
        if abruptvars is None:
            self.abruptvars = self.DEFAULT_ABRUPT_VARS
        elif is_empty(abruptvars):
            self.abruptvars = {}
        else:
            if not isinstance(abruptvars, Mapping):
                raise ValueError("Only Mapping (dict) is expected.")
            self.abruptvars = abruptvars

    def __init__(self, *args, **kwargs):
        subject = kwargs.pop('subject')
        capacity = kwargs.pop('capacity', None)
        flushLevel = kwargs.pop('flushLevel', FINISHED)
        target = kwargs.pop('target', None)
        flushOnClose = kwargs.pop('flushOnClose', True)

        super().__init__(*args, **{**kwargs,
                                   'subject': subject, 'capacity': capacity, 'flushLevel': flushLevel,
                                   'target': target,
                                   'flushOnClose': flushOnClose})
        self.calc_abrupt_vars(*args, **kwargs)



@contextlib.contextmanager
def EmailStatus(emailLogger, logger=None, faildargs={}, successargs={}, successkwargs={}, faildkwargs={}):
    """
    if contextmanager exits with exception (it fails), than e-mail with subject formatted with faildargs and faildkwargs
    will be send.
    Otherwise, e-mail with subject formatted with successargs and successkwargs will be send.

    All messages (in the current Thread) will be aggregated to one long e-mail with the subject described in
    OneMemoryHandler.get_subject() method. (see also below).

    You can have subject of the e-mails without place holders, for example, 'Aggregates log from the Demo application'.

    Below, is example of customizable e-mails message.

    Indented usage example:

    with EmailStatus(emailLogger=emailLogger, logger=logger, faildargs={'status': 'Failed'},
                              successargs={'status': 'Done'}):
        emailLogger.info("Start - Process")
        ...
        emailLogger.info("Step 1")
        ...
        emailLogger.info("Step 2")
        ...
        emailLogger.info("Done Successfully - Process")

    'status' is place holder in the subject of the e-mail. For example, 'My Process Status : {{status}}'.
    In this case the same subject of the e-mail will be used whether the code-block finish successfully or with failure.

    If you want to change delimeters used to indicate variable declaration inside template, see docstring of the
    OneMemoryHandler.get_subject() method.

    :param emailLogger: configured emailLogger. See https://docs.python.org/3/howto/logging-cookbook.html#logging-cookbook
    :param faildargs:   variable resolution for e-mail's subject on failure. Optional.
                            Expected to be dict with variable substitution.
    :param successargs: variable resolution for e-mail's subject on success. Optional.
                            Expected to be dict with variable substitution.
    :param logger:      logger ot used if exception was thrown from the code. It will be used in addition to
                        emailLogger. Optional.
    :param successkwargs: to be send to the emailLogger's kwargs param on success. Optional.
    :param faildkwargs:   to be send to the emailLogger's kwargs param on failure. Optional.
    :return:
    """

    try:
        yield emailLogger
    except Exception:
        if logger is not None:
            logger.error("", exc_info=True)
        emailLogger.error("", exc_info=True)
        emailLogger.log(FINISHED, faildargs, **faildkwargs)
    else:
        emailLogger.log(FINISHED, successargs, **successkwargs)

class NoDefaultSmtpPortWarning(RuntimeWarning):
    pass


def initConfig(**kwargs):
    """
    This method can be optionally called prior any call to another function in this module.
    It is indented to be called in the MainThread.
    This method can be call with empty params.

    By default, SMTP class from smtplib is used to send actual e-mail. You can change it to SMTP_SSL, LMTP,
    or another class by specifying default_smtp_cls_name.

    SMTP and some other classes (but not all) has default_port class-level field. If you use
    class that doesn't have such field you have 2 options:
        1.Specify port with default_smtp_port param.
        2.Explicitly use port in mailhost param to SMTPHandler as list of mailhost and port.

    In any case, the order of the determination of the port is as following:

    1. Port in mailhost param to SMTPHandler.
    2. default_smtp_cls_name if exists
    3. default_smtp_port

    This method is idempotent.

    :param default_smtp_cls: Can be class or str. Optional
            Default values is: 'smtplib.SMTP'

    :param default_smtp_port: Optional
                Default values is: `smtplib.SMTP.default_port` if exists else None.
    :return:
    """

    default_smtp_cls_p = kwargs.get('default_smtp_cls_name', None)
    if default_smtp_cls_p is None:
        default_smtp_cls_p = _SMTP
    elif isinstance(default_smtp_cls_p, str):
        default_smtp_cls_p = importer(default_smtp_cls_p)
    global default_smtp_cls
    default_smtp_cls = default_smtp_cls_p

    default_smtp_port_p = kwargs.get('default_smtp_port', None)
    if default_smtp_port_p is None:
        default_smtp_port_p = getattr(_SMTP, 'default_port', None)
    elif isinstance(default_smtp_port_p, str):
        default_smtp_port_p = importer(default_smtp_port_p)
    global default_smtp_port
    default_smtp_port = default_smtp_port_p




initConfig()


"""
This module is usable in your deployment script. See also `deploys` module.

This module depends on some 3-rd party dependencies, in order to use it, you should have installed first. To do it
run `python3 -m pip install alex-ber-utils[fabric]`.

"""


import logging
logger = logging.getLogger(__name__)
from pathlib import PurePath

try:
    from fabric import Connection
except ImportError:
    import warnings

    warning = (
        "You appear to be missing some optional dependencies;"
        "please 'python3 -m pip install alex-ber-utils[fabric]'."
    )
    warnings.warn(warning, ImportWarning)
    raise




def _cp(self, remote_dir, local_file):
    """
    This is helper method that cp single file from local machine to remote (Posix) machine.
    If file exists on the remote machine, it will be overwritten.

    :param remote_dir: can be str or Path
    :param local_file:  can be str or Path
    """
    remote_dir_path = PurePath(remote_dir)
    self.run(f'mkdir -p {remote_dir_path.as_posix()}')
    #self.run('ls -l {}'.format(remote_dir))
    full_local_dir = PurePath(local_file)
    file_name = full_local_dir.name
    local_dir = full_local_dir.parent

    full_remote_dir = remote_dir_path / file_name
    self.run(f'rm -fr {full_remote_dir.as_posix()}')
    full_local_dir = PurePath(local_dir, file_name)
    result = self.put(local=full_local_dir.as_posix(), remote=remote_dir_path.as_posix())
    logger.info(f"Uploaded {result.local} to {result.remote}")

Connection.cp = _cp


import enum
import logging

import pytest

from alexber.utils.enums import Enum
from alexber.utils.parsers import ConfigParser, ArgumentParser, safe_eval, is_empty, parse_boolean, parse_sys_args

logger = logging.getLogger(__name__)
from decimal import Decimal
from datetime import datetime
from importlib.resources import path

def test_parse_config(request):
    logger.info(f'{request._pyfuncitem.name}()')

    parser = ConfigParser()

    pck = '.'.join(['tests_data', __package__, 'parser'])

    with path(pck, 'config.ini') as f:
        parser.read(f)

    dd = parser.as_dict()
    da = dd['playera']
    clsa = da['cls']
    namea = da['name']

    db = dd['playerb']
    clsb = db['cls']
    nameb = db['name']

    assert clsa==clsb
    assert namea == nameb


@pytest.fixture(params=[
    #'args,exp_d',
    ('--key=value --single',
     dict([('key', 'value'), ('single', None)])),

    ('wrong=pair',
         dict([('wrong=pair', None)])),

    ('--conf --prop1=value1 --prop2=value --prop1=value9',
         dict([('conf', None), ('prop1', 'value9'), ('prop2', 'value')])),

])
def arg_parse_param(request):
    return request.param

def test_args_parse(request, mocker, arg_parse_param):
    logger.info(f'{request._pyfuncitem.name}()')
    args, exp_d = arg_parse_param

    parser = ArgumentParser()

    mock_args = mocker.patch('alexber.utils.parsers.sys.argv', new_callable=list)
    mock_args.append(__file__)
    mock_args[1:] = args.split()

    d = parser.as_dict()
    assert exp_d==d


def test_args_parse_explicit_args(request, arg_parse_param):
    logger.info(f'{request._pyfuncitem.name}()')
    args, exp_d = arg_parse_param

    parser = ArgumentParser()

    sys_args = args.split()

    d = parser.as_dict(args=sys_args)
    assert exp_d==d


@pytest.mark.parametrize(
     'value, exp_value, exp_type',

    [
        ('John', 'John', str),
        ('alexber.utils.players.ConstantPlayer', 'alexber.utils.players.ConstantPlayer', str),
        ('1000', 1000, int),
        ('0.1', 0.1, float),
        ('0.0', 0.0, float),
        ('-0.0', -0.0, float),
        ('-5', -5, int),
        ('0.1', None, Decimal),  #Not Supprted
        ('2019-04-01 16:31:51.513383', None, datetime),  #Not Supprted
        ('%(message)s','%(message)s', str),#https://github.com/alex-ber/AlexBerUtils/issues/2
        ('(message)s','(message)s', str),
        ('(message)','(message)', str),

    ]
)
def test_convert(request, value, exp_value, exp_type):
    logger.info(f'{request._pyfuncitem.name}()')
    if exp_value is None:
        logger.debug(f"Type {exp_type} is not supported.")
        return


    result = safe_eval(value)
    type_result = type(result)
    pytest.assume(exp_value == result)
    pytest.assume(exp_type == type_result)


@enum.unique
class Color(Enum):
    RED = 'r'
    BLUE = 'b'
    GREEN = 'g'


@pytest.mark.parametrize(
    'value, exp_result',
    [
     (True, False),
     (False, True),
     (None, True),

     ("something", False),
     ("", True),
     #
     (1, False),
     (0, True),
     (0.0, True),

     ("1", False),
     ("0", False),

     (Color.RED, False),
     ([], True),
     ([None], False),
     (['something'], False),
     ]
)
def test_is_empty(request, value, exp_result):
    logger.info(f'{request._pyfuncitem.name}()')

    result = is_empty(value)
    assert exp_result == result


@pytest.mark.parametrize(
    'value, exp_result',
    [
     (True, True),
     (False, False),
     (None, None),

     ("True", True),
     ("False", False),

     ("TRUE", True),
     ("FALSE", False),

     ("tRuE", True),
     ("fALsE", False),

     ("true", True),
     ("false", False),

     (1, True),
     (0, False),
     (0.0, False),
     ]
)
def test_parse_boolean(request, value, exp_result):
    logger.info(f'{request._pyfuncitem.name}()')

    result = parse_boolean(value)
    assert exp_result == result

@pytest.mark.parametrize(
    'value',
    [
     ("gibrish123"),
     ("T"),
     ("F"),

     ("t"),
     ("f"),

     ("1"),
     ("0"),

     (3.5),
     ([]),
     (5),
     (2.01),

    ]
)


def test_parse_boolean_invalid(request, value):
    logger.info(f'{request._pyfuncitem.name}()')

    with pytest.raises(ValueError, match='nknown'):
        parse_boolean(value)



def test_parse_sys_args(request):
    logger.info(f'{request._pyfuncitem.name}()')
    expdd = {
        'general.profiles': "dev",
        'general.log.formatters.detail.format': '%(message)s',
        'general.log.root.level': '20', #logging.INFO
        'app.inner_host_name': 'yahoo.com',
        'app.white_list': 'gamma,alpha,betha',
        'app.alt_white_list': '100,10.0'
    }

    argsv = '--general.profiles=dev ' \
            '--general.log.formatters.detail.format=%(message)s ' \
            '--general.log.root.level=20 ' \
            '--app.inner_host_name=yahoo.com ' \
            '--app.white_list=gamma,alpha,betha ' \
            '--app.alt_white_list=100,10.0 ' \
        .split()


    _, dd = parse_sys_args(args=argsv)
    assert expdd == dd


if __name__ == "__main__":
    pytest.main([__file__])



#see https://stackoverflow.com/questions/1785503/when-should-i-use-uuid-uuid1-vs-uuid-uuid4-in-python
#https://stackoverflow.com/questions/703035/when-are-you-truly-forced-to-use-uuid-as-part-of-the-design/786541#786541

from uuid import uuid1 as _uuid1
_int_from_bytes = int.from_bytes  # py3 only

from random import SystemRandom as _SystemRandom

_system_random = _SystemRandom()





def uuid1mc():
    '''
    This is v1 with random MAC ("v1mc"). This is a hybrid between version 1 & version 4.

    Version 1 UUIDs use the network card's MAC address (which unless spoofed, should be unique),
    plus a timestamp, plus the usual bit-twiddling to generate the UUID.

    Version 4 UUIDs Generate a random UUID.

    uuid1mc() is deliberately generating v1 UUIDs with a random broadcast MAC address (this is allowed by the v1 spec).
    The resulting v1 UUID is time dependant (like regular v1), but lacks all host-specific information (like v4).
    It's also much closer to v4 in it's collision-resistance:
        v1mc = 60 bits of time + 61 random bits = 121 unique bits; v4 = 122 random bits.

    Note: somebody reported that ran into trouble using UUID1 in Amazon EC2 instances.
    He use UUID1 for a database upgrade script where he generated ~120k UUIDs within a couple of minutes.
    The UUID collision led to violation of a primary key constraint.
    He suspect poor clock resolution and switching to UUID4 solved it for him.

    '''
    #return uuid1(_int_from_bytes(urandom(6), "big") | 0x010000000000)
    node = _system_random.getrandbits(8) #6 and not 8, because this function round up to bits / 8 and rounded up
    # NOTE: The constant here is required by the UUIDv1 spec...
    return  _uuid1(node | 0x010000000000)





"""
This is essentially wrapper arround HiYaPyCo project with streamlined and extended API and couple of work-arrounds.

This module depends on some 3-rd party dependencies, in order to use it, you should have installed first. To do it
run `python3 -m pip install alex-ber-utils[yml]`.

This module doesn't use any package-level variables in hiYaPyCo module, including hiYaPyCo.jinja2env.
This module do use Jinja2's `Environment`.

This module is low level API for init_app_conf.py, deploys.py. It is also optionally used in emails.py.

See here https://medium.com/analytics-vidhya/my-ymlparsers-module-88221edf16a6 for documentation.

"""

import warnings
try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=r'.*?collections\.abc.*?', module=r'.*?jinja2.*?')
        from jinja2 import Environment as _Environment, \
                        DebugUndefined as _DebugUndefined
except ImportError:
    import warnings

    warning = (
        "You appear to be missing some optional dependencies (jinja2);"
        "please 'python3 -m pip install alex-ber-utils[yml]'."
    )
    warnings.warn(warning, ImportWarning)
    raise

try:
    import hiyapyco as _hiyapyco
except ImportError:
    import warnings

    warning = (
        "You appear to be missing some optional dependencies (hiyapyco);"
        "please 'python3 -m pip install alex-ber-utils[yml]'."
    )
    warnings.warn(warning, ImportWarning)
    raise

try:
    _hiyapyco.METHOD_SUBSTITUTE
except AttributeError as e:
    import warnings

    warning = (
        "You appear to be missing some optional yml parsing dependencies (hiyapyco should be at least 0.4.16);"
        "please 'python3 -m pip install alex-ber-utils[yml]'."
    )
    warnings.warn(warning, ImportWarning)
    raise ImportError(str(e)) from e

import io as _io
import os as _os
import contextlib
from threading import RLock as _Lock


from hiyapyco import logger, HiYaPyCoImplementationException, HiYaPyCoInvocationException, TemplateError, \
    re, strtobool, HiYaPyCo as _HiYaPyCo
from hiyapyco.odyldo import safe_dump as _safe_dump

_safe_dump_d = None
_load_d = None

from platform import uname as _uname
from . _ymlparsers_extra import format_template as _format_template

format_template = _format_template

class HiYaPyCo(_HiYaPyCo):
    jinja2ctx = None
    jinja2Lock = None


    #this adopted version
    #1. In order to change reference of jinja2env
    #2. In order to use reentrant lock
    def _interpolatestr(self, s):
        try:
            with self.jinja2Lock:
                si = HiYaPyCo.jinja2ctx.from_string(s).render(self._data)
        except TemplateError as e:
            # FIXME: this seems to be broken for unicode str?
            raise HiYaPyCoImplementationException('error interpolating string "%s" : %s' % (s, e,))
        if not s == si:
            if self.castinterpolated:
                if not re.match( r'^\d+\.*\d*$', si):
                    try:
                        si = bool(strtobool(si))
                    except ValueError:
                        pass
                else:
                    try:
                        if '.' in si:
                            si = float(si)
                        else:
                            si = int(si)
                    except ValueError:
                        pass
            logger.debug('interpolated "%s" to "%s" (type: %s)' % (s, si, type(si),))
        return si


def safe_dump(data, stream=None, **kwds):
    """
    Dumping data to stream.

    Simple Python objects like primitive types (str, integer, etc), list, dict, OrderedDict are supported.

    Note: that collections.OrderedDict is also supported.

    :param data: data-structure to dump.
    :param stream: where to write the representation.
    :param kwds: See initConfig() for default values. See yaml.dump_all()
    :return: str representation of data
    """
    if _safe_dump_d is None:
        raise ValueError("_safe_dump_d can't be None")

    kwargs = {**_safe_dump_d, **kwds}
    _safe_dump(data, stream, **kwargs)


def load(*args, **kwds):
    """
    Load a Hierarchical yml files
    --------------------------------------

    See initConfig() for default values.
    If you want to disable variable substitution, use DisableVarSubst context manager like this:

    with ymlparsers.DisableVarSubst():
        d = ymlparsers.load([str(full_path)])

    :param args: YMLfile(s)
    :param kwargs:
      * method: one of hiyapyco.METHOD_SIMPLE | hiyapyco.METHOD_MERGE | hiyapyco.METHOD_SUBSTITUTE
      * mergelists: boolean (default: True) try to merge lists (only makes sense if hiyapyco.METHOD_MERGE or hiyapyco.METHOD_SUBSTITUTE)
      * interpolate: boolean (default: False)
      * castinterpolated: boolean (default: False) try to cast values after interpolating
      * encoding: (default: 'utf-8') encoding used to read yml files
      * loglevel: one of  the valid levels from the logging module
      * failonmissingfiles: boolean (default: True)
      * loglevelmissingfiles

    :returns a representation of the merged and (if requested) interpolated config using OrderedDict.
    """
    if _load_d is None:
        raise ValueError("_load_d can't be None")

    kwargs = {**_load_d, **kwds}
    #return _hiyapyco.load(*args, **kwargs)
    hiyapyco = HiYaPyCo(*args, **kwargs)
    return hiyapyco.data()


def as_str(data, **kwds):
    """
    Return str representation of the data.

    Simple Python objects like primitive types (str, integer, etc), list, dict, OrderedDict are supported.

    Note: that collections.OrderedDict is also supported.

    :param data: data-structure to dump.
    :param kwds: See initConfig() for default values. See yaml.dump_all()
    :return: str representation of data
    """
    with _io.StringIO() as buf:
        buf.write(_os.linesep)
        safe_dump(data, stream=buf, **kwds)
        return buf.getvalue()

@contextlib.contextmanager
def DisableVarSubst(*args, **kwargs):
    """
    Use of this context manager disables variable substation in the load() function.


    :param jinja2ctx - Jinja2 Environment. If not provided HiYaPyCo.jinja2ctx is used.
    :param jinja2Lock - lock to use for synchronization. Should be the same here and in load() function.
                        If not provided HiYaPyCo.jinja2ctx is used.
    """
    jinja2ctx = kwargs.pop('jinja2ctx', None)
    jinja2Lock = kwargs.pop('jinja2Lock', None)

    if jinja2ctx is None and jinja2Lock is not None:
        raise ValueError("You can't provide your jinja2Lock for HiYaPyCo.jinja2ctx")

    if HiYaPyCo.jinja2ctx is not None and HiYaPyCo.jinja2Lock is None:
        raise ValueError('HiYaPyCo.jinja2ctx is not None, but HiYaPyCo.jinja2Lock is None')

    if jinja2ctx is None:
        jinja2ctx = HiYaPyCo.jinja2ctx
        jinja2Lock = HiYaPyCo.jinja2Lock

    if jinja2ctx is None or jinja2Lock is None:
        raise ValueError("You should pass both jinja2ctx and jinja2Lock or "
                         "You should have HiYaPyCo.jinja2ctx and HiYaPyCo.jinja2Lock not None"
                         )

    with jinja2Lock:
        variable_start_string = jinja2ctx.variable_start_string
        variable_end_string = jinja2ctx.variable_end_string
        block_start_string = jinja2ctx.block_start_string
        block_end_string = jinja2ctx.block_end_string

        jinja2ctx.variable_start_string = '@@{@|'
        jinja2ctx.variable_end_string = '|@}@@'
        jinja2ctx.block_start_string = '%%{%|'
        jinja2ctx.block_end_string = '|%}%%'

        try:
            yield jinja2ctx
        finally:
            jinja2ctx.variable_start_string = variable_start_string
            jinja2ctx.variable_end_string = variable_end_string
            jinja2ctx.block_start_string = block_start_string
            jinja2ctx.block_end_string = block_end_string


# # support for tag 'tag:yaml.org,2002:python/object/apply:collections.OrderedDict'
# _represent_dict_order = lambda self, data:  self.represent_mapping('tag:yaml.org,2002:map', data.items())
# yaml.add_representer(OrderedDict, _represent_dict_order)

#
# _hiyapyco.load("\n") #_hiyapyco._usedefaultyamlloader = False



def initConfig(**kwargs):
    """
    This method can be optionally called prior any call to another function in this module.
    It is indented to be called in the MainThread.
    This method can be call with empty params.

    Note: this module doesn't use any package-level variables in hiYaPyCo module.

    :param jinja2Lock: lock to use for synchronization in DisableVarSubst and load().
                      If not supplied, default RLock is used.
                      Will be available as HiYaPyCo.jinja2Lock.
    :param jinja2ctx: context for overriding values in initialization (default is _DebugUndefined) of Jinja2 Environment:
                      gloabls will be override values in Environment.globals.update(**globals)
                      Default is 'uname':platform.uname
                      Will be available as HiYaPyCo.jinja2ctx.
    :param load: this params will be used as default values in load() function. See hiyapyco.load()
                Default values are
                      'method':_hiyapyco.METHOD_SUBSTITUTE,
                      'mergelists':False,
                      'interpolate':True,
                      'castinterpolated':True
                This means, by default:
                      We're replacing list, not merging them.
                      We're interpolating values in the data (to scalar/list/OrderDict).
                      We're also using casting to appropriate type.
    :param safe_dump: this params will be used as default values in safe_dump() and as_str(). See yaml.dump_all()
                Default values are
                      'default_flow_style':False,
                      'sort_keys':False
                This means, by default:
                       we prefer block style always.
                       we preserve the key order (no sorting for key in the dictionary).

    If running from the MainThread, this method is idempotent.

    :return:
    """
    jinja2Lock_p = kwargs.get('jinja2Lock', None)
    if jinja2Lock_p is None:
        jinja2Lock_p = _Lock()
    HiYaPyCo.jinja2Lock = jinja2Lock_p

    jinja2ctx_d = kwargs.get('jinja2ctx', {})
    globals_d = jinja2ctx_d.pop('globals', {})
    # jinja2ctx_p = _Environment(undefined=_StrictUndefined)
    jinja2ctx_p = _Environment(**{'undefined':_DebugUndefined, **jinja2ctx_d})
    jinja2ctx_p.globals.update(**{'uname':_uname, **globals_d})


    with HiYaPyCo.jinja2Lock:
        HiYaPyCo.jinja2ctx = jinja2ctx_p


    load_d = kwargs.get('load', {})
    _load_d_p = {'method':_hiyapyco.METHOD_SUBSTITUTE,
               'mergelists':False,
               'interpolate':True,
               'castinterpolated':True,
               **load_d}

    method = _load_d_p['method']
    # TODO: HiYaPyCo._substmerge() bug workarround, see https://github.com/zerwes/hiyapyco/pull/38
    HiYaPyCo._deepmerge = _HiYaPyCo._substmerge
    if method == _hiyapyco.METHOD_MERGE:
        #restore original _deepmerge
        HiYaPyCo._deepmerge = _HiYaPyCo._deepmerge
    global _load_d
    _load_d = _load_d_p


    safe_dump_d = kwargs.get('safe_dump', {})
    global _safe_dump_d
    #See https://github.com/yaml/pyyaml/pull/256
    _safe_dump_d = {'default_flow_style':False,
                        'sort_keys':False,
                        **safe_dump_d}



initConfig()


import logging
import pytest
from importlib.resources import path
from pathlib import Path

from alexber.utils.deploys import split_path, load_config, add_to_zip_copy_function
from tests.utils.ymlparsers_test import ymlparsersSetup, ymlparsersCleanup, exp_config_d
from tests.utils.init_app_conf_test import initappconfFixture
import copy
import tempfile
import zipfile
import shutil
try:
    import alexber.utils.ymlparsers as ymlparsers
except ImportError:
    pass
logger = logging.getLogger(__name__)

class TestSplitPath(object):
    def test_split_path_intented(self, request):
        logger.info(f'{request._pyfuncitem.name}()')
        split_dirname = 'ymlparsers'
        pck = '.'.join(['tests_data', __package__, split_dirname])

        with path(pck, 'config.yml') as f:
            first_part, second_part = split_path(f, split_dirname)
            pytest.assume(f.parent==first_part)
            pytest.assume('config.yml'==str(second_part))

    def test_split_path_intented_path(self, request):
        logger.info(f'{request._pyfuncitem.name}()')
        split_dirname = 'ymlparsers'
        pck = '.'.join(['tests_data', __package__, split_dirname])

        with path(pck, 'config.yml') as f:
            first_part, second_part = split_path(Path(f), Path(split_dirname))
            pytest.assume(f.parent==first_part)
            pytest.assume('config.yml'==str(second_part))



    def test_split_path_absent_undefined_behaviour(self, request):
        logger.info(f'{request._pyfuncitem.name}()')
        pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

        with path(pck, 'config.yml') as f:
            with pytest.raises(BaseException):
                split_path(f, 'xxxnotfoundxxx')


    def test_split_path_absent_compount_name(self, request):
        logger.info(f'{request._pyfuncitem.name}()')
        pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

        with path(pck, 'config.yml') as f:
            with pytest.raises(BaseException):
                split_path(f, str(Path('ymlparsers', 'config.yml')))

    def test_split_path_None(self, request):
        logger.info(f'{request._pyfuncitem.name}()')
        split_dirname = 'ymlparsers'
        pck = '.'.join(['tests_data', __package__, split_dirname])

        with path(pck, 'config.yml') as f:
            with pytest.raises(BaseException, match='None'):
                split_path(f, None)

    def test_split_path_empty_str(self, request):
        logger.info(f'{request._pyfuncitem.name}()')
        split_dirname = 'ymlparsers'
        pck = '.'.join(['tests_data', __package__, split_dirname])

        with path(pck, 'config.yml') as f:
            with pytest.raises(BaseException):
                split_path(f, None)

    def test_split_path_twice_split_dir(self, request):
        logger.info(f'{request._pyfuncitem.name}()')
        split_dirname = 'tests'
        pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

        with path(pck, 'config.yml') as f:
            assert split_dirname in f.parts
            duplicate_dir_base = f.parent / split_dirname
            duplicate_dir_full_path = duplicate_dir_base / f.name
            first_part, second_part = split_path(duplicate_dir_full_path, split_dirname)

            pytest.assume(duplicate_dir_base==first_part)
            pytest.assume('config.yml'==str(second_part))

def check_config_yml(default_d, exp_d):


    app_d = default_d.get('app', None)
    exp_app_d = exp_d.get('app', None)

    inner_host_name = app_d.get('inner_host_name', None)
    exp_host_name = exp_app_d.get('inner_host_name', None)
    pytest.assume(exp_host_name==inner_host_name)
    cli_template = app_d.get('cli_template')
    pytest.assume('inner_host_name' in cli_template)
    pytest.assume(exp_d == default_d)

@pytest.mark.yml
def test_load_config(request, mocker, ymlparsersSetup, ymlparsersCleanup, initappconfFixture, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')

    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])
    exp_profiles=['dev']
    with path(pck, 'config.yml') as full_path:
        args = f"--general.config.file={full_path} --general.profiles={exp_profiles[0]}".split()
        config_path, default_d = load_config(None, args)

    pytest.assume(full_path == config_path)
    general_d = default_d.get('general', None)
    profiles = general_d.get('profiles', None)
    pytest.assume(exp_profiles == profiles)

    exp_d = copy.deepcopy(exp_config_d)
    exp_d['general']['profiles'] = exp_profiles

    check_config_yml(default_d, exp_d)

@pytest.mark.yml
def test_add_to_zip_copy_function(request, mocker,  ymlparsersSetup, ymlparsersCleanup, exp_config_d):
    logger.info(f'{request._pyfuncitem.name}()')
    split_dirname = 'ymlparsers'
    pck = '.'.join(['tests_data', __package__, 'ymlparsers'])

    with path(pck, 'config.yml') as full_path:


        with tempfile.TemporaryDirectory() as temp_dir:
            local_folder = Path(temp_dir)
            zip_path = local_folder / "my.zip"
            with zipfile.ZipFile(zip_path, mode="w") as zf:
                # not real copy, add to zip archive
                shutil.copytree(str(full_path.parent), str(local_folder / "xxxirrelevantxxx"),
                                ignore=shutil.ignore_patterns('__pycache__', 'venv', 'logs', '*.pyc', 'tmp*',
                                                              '.*', "__init__.py"),
                                copy_function=add_to_zip_copy_function(split_dirname=split_dirname, zf=zf),
                                symlinks=True, ignore_dangling_symlinks=True)
            with zipfile.ZipFile(zip_path, mode="r") as zf:
                entry_path = (Path(split_dirname) / 'config.yml').as_posix()
                entry_content_bytes = zf.read(str(entry_path))

                with ymlparsers.DisableVarSubst():
                    default_d = ymlparsers.load([entry_content_bytes.decode('utf-8')])

        check_config_yml(default_d, exp_config_d)



if __name__ == "__main__":
    pytest.main([__file__])







"""
A Python replacement for java.util.Properties class
This is modelled as closely as possible to the Java original.

Created - Anand B Pillai <abpillai@gmail.com>,
Update to Python 3 by Alex.
Also there are some tweeks that was done by Alex.
"""

# see https://code.activestate.com/recipes/496795-a-python-replacement-for-javautilproperties/
import sys
import re
import time
from collections import OrderedDict


class IllegalArgumentException(Exception):

    def __init__(self, lineno, msg):
        self.lineno = lineno
        self.msg = msg

    def __str__(self):
        s='Exception at line number %d => %s' % (self.lineno, self.msg)
        return s
                 
class Properties(object):
    """ A Python replacement for java.util.Properties """
    # Alex fix, handle ${}
    #curlies = re.compile('\\${.+?}')
    PLACEHOLDER_TOKEN = '$'

    def __init__(self, props=None):

        # Note: We don't take a default properties object
        # as argument yet

        # Dictionary of properties.
        # Alex changed {} to OrderedDict
        self._props = OrderedDict()
        # Dictionary of properties with 'pristine' keys
        # This is used for dumping the properties to a file
        # using the 'store' method
        # Alex changed {} to OrderedDict
        self._origprops = OrderedDict()

        # Dictionary mapping keys from property
        # dictionary to pristine dictionary
        # Alex changed {} to OrderedDict
        self._keymap = OrderedDict()
        # Alex changed {} to OrderedDict
        self._unresolved = OrderedDict()
        
        self.othercharre = re.compile(r'(?<!\\)(\s*\=)|(?<!\\)(\s*\:)')
        self.othercharre2 = re.compile(r'(\s*\=)|(\s*\:)')
        self.bspacere = re.compile(r'\\(?!\s$)')
        
    def __str__(self):
        s='{'
        for key,value in self._props.items():
            s = ''.join((s,key,'=',value,', '))

        s=''.join((s[:-2],'}'))
        return s

    def __parse(self, lines):
        """ Parse a list of lines and create
        an internal property dictionary """

        # Every line in the file must consist of either a comment
        # or a key-value pair. A key-value pair is a line consisting
        # of a key which is a combination of non-white space characters
        # The separator character between key-value pairs is a '=',
        # ':' or a whitespace character not including the newline.
        # If the '=' or ':' characters are found, in the line, even
        # keys containing whitespace chars are allowed.

        # A line with only a key according to the rules above is also
        # fine. In such case, the value is considered as the empty string.
         # In order to include characters '=' or ':' in a key or value,
        # they have to be properly escaped using the backslash character.

        # Some examples of valid key-value pairs:
        #
        # key     value
        # key=value
        # key:value
        # key     value1,value2,value3
        # key     value1,value2,value3 \
        #         value4, value5
        # key
        # This key= this value
        # key = value1 value2 value3
        
        # Any line that starts with a '#' is considerered a comment
        # and skipped. Also any trailing or preceding whitespaces
        # are removed from the key/value.
        
        # This is a line parser. It parses the
        # contents like by line.

        lineno=0
        i = iter(lines)

        for line in i:
            lineno += 1
            line = line.strip()
            # Skip null lines
            if not line: continue
            # Skip lines which are comments
            if line[0] == '#': continue
            # Some flags
            escaped=False
            # Position of first separation char
            sepidx = -1
            # A flag for performing wspace re check
            flag = 0
            # Check for valid space separation
            # First obtain the max index to which we
            # can search.
            m = self.othercharre.search(line)
            if m:
                first, last = m.span()
                start, end = 0, first
                flag = 1
                wspacere = re.compile(r'(?<![\\\=\:])(\s)')        
            else:
                if self.othercharre2.search(line):
                    # Check if either '=' or ':' is present
                    # in the line. If they are then it means
                    # they are preceded by a backslash.
                    
                    # This means, we need to modify the
                    # wspacere a bit, not to look for
                    # : or = characters.
                    wspacere = re.compile(r'(?<![\\])(\s)')        
                start, end = 0, len(line)
                
            m2 = wspacere.search(line, start, end)
            if m2:
                # print 'Space match=>',line
                # Means we need to split by space.
                first, last = m2.span()
                sepidx = first
            elif m:
                # print 'Other match=>',line
                # No matching wspace char found, need
                # to split by either '=' or ':'
                first, last = m.span()
                sepidx = last - 1
                # print line[sepidx]
                
                
            # If the last character is a backslash
            # it has to be preceded by a space in which
            # case the next line is read as part of the
            # same property
            while line[-1] == '\\':
                # Read next line
                nextline = i.next()
                nextline = nextline.strip()
                lineno += 1
                # This line will become part of the value
                line = line[:-1] + nextline

            # Now split to key,value according to separation char
            if sepidx != -1:
                key, value = line[:sepidx], line[sepidx+1:]
            else:
                key,value = line,''

            self.processPair(key, value)
            
    def processPair(self, key, value):
        """ Process a (key, value) pair """

        oldkey = key
        oldvalue = value
        
        # Create key intelligently
        keyparts = self.bspacere.split(key)
        # print keyparts

        strippable = False
        lastpart = keyparts[-1]

        if lastpart.find('\\ ') != -1:
            keyparts[-1] = lastpart.replace('\\','')

        # If no backspace is found at the end, but empty
        # space is found, strip it
        elif lastpart and lastpart[-1] == ' ':
            strippable = True

        key = ''.join(keyparts)
        if strippable:
            key = key.strip()
            oldkey = oldkey.strip()
        
        oldvalue = self.unescape(oldvalue)
        value = self.unescape(value)
        
        # Alex fix, handle ${}
        _regex ='\\${.+?}'
        if '$' != Properties.PLACEHOLDER_TOKEN:
            _regex = _regex[0:1] + Properties.PLACEHOLDER_TOKEN + _regex[2:]
        curlies = re.compile(_regex)
        found = curlies.findall(value)
        
        for f in found:
            srcKey = f[2:-1]
            if srcKey in self._props:
                value = value.replace(f, self._props[srcKey], 2)
        
        self._props[key] = value.strip()
        found = curlies.findall(value.strip())
        
        for f in found:
            srcKey = f[2:-1]
            if self._unresolved.has_key(srcKey):
                self._unresolved[srcKey] += [key]
            else:
                self._unresolved[srcKey] = [key]

        #End of Alex fix
        
        # Check if an entry exists in pristine keys
        if key in self._keymap:
            oldkey = self._keymap.get(key)
            self._origprops[oldkey] = oldvalue.strip()
        else:
            self._origprops[oldkey] = oldvalue.strip()
            # Store entry in keymap
            self._keymap[key] = oldkey
            
         # Alex fix, handle ${}
        if key in self._unresolved:
            for updateKey in self._unresolved[key]:
                self.processPair(updateKey, self._props[updateKey])
        
    def escape(self, value):

        # Java escapes the '=' and ':' in the value
        # string with backslashes in the store method.
        # So let us do the same.
        #Alex fix added r prefix to str
        newvalue = value.replace(':',r'\:')
        # Alex fix added r prefix to str
        newvalue = newvalue.replace('=',r'\=')

        return newvalue

    def unescape(self, value):

        # Reverse of escape
        # Alex fix added r prefix to str
        newvalue = value.replace(r'\:',':')
        # Alex fix added r prefix to str
        newvalue = newvalue.replace(r'\=','=')

        return newvalue    
        
    def load(self, stream):
        """ Load properties from an open file stream """

        #Alex disabled unneccesery stricked validations
        # For the time being only accept file input streams
        #if type(stream) is not file:
        #    raise TypeError('Argument should be a file object!')
        ## Check for the opened mode
        #if stream.mode != 'r':
        #    raise ValueError('Stream should be opened in read-only mode!')

        try:
            lines = stream.readlines()
            self.__parse(lines)
        except IOError as e:
            raise

    def getProperty(self, key):
        """ Return a property for the given key """
        
        return self._props.get(key,'')

    def setProperty(self, key, value):
        """ Set the property for the given key """

        if type(key) is str and type(value) is str:
            self.processPair(key, value)
        else:
            raise TypeError('both key and value should be strings!')

    def propertyNames(self):
        """ Return an iterator over all the keys of the property
        dictionary, i.e the names of the properties """

        return self._props.keys()

    def list(self, out=sys.stdout):
        """ Prints a listing of the properties to the
        stream 'out' which defaults to the standard output """

        out.write('-- listing properties --\n')
        for key,value in self._props.items():
            out.write(''.join((key,'=',value,'\n')))

    def store(self, out, header=""):
        """ Write the properties list to the stream 'out' along
        with the optional 'header' """

        if out.mode[0] != 'w':
            raise ValueError('Steam should be opened in write mode!')

        try:
            out.write(''.join(('#',header,'\n')))
            # Write timestamp
            tstamp = time.strftime('%a %b %d %H:%M:%S %Z %Y', time.localtime())
            out.write(''.join(('#',tstamp,'\n')))
            # Write properties from the pristine dictionary
            for prop, val in self._origprops.items():
                out.write(''.join((prop,'=',self.escape(val),'\n')))
                
            out.close()
        except IOError as e:
            raise

    # Alex changed name getPropertyDict() to current
    def as_dict(self):
        return self._props

    def __getitem__(self, name):
        """ To support direct dictionary like access """

        return self.getProperty(name)

    def __setitem__(self, name, value):
        """ To support direct dictionary like access """

        self.setProperty(name, value)
        
    def __getattr__(self, name):
        """ For attributes not found in self, redirect
        to the properties dictionary """

        try:
            return self.__dict__[name]
        except KeyError:
            if hasattr(self._props,name):
                return getattr(self._props, name)

